"""Qt-Specific GUI addons to dancer"""
from PySide6 import QtCore as _QtCore, QtGui as _QtGui, QtWidgets as _QtWidgets
from PySide6.QtWidgets import (QMessageBox as _QMessageBox, QCheckBox as _QCheckBox, QBoxLayout as _QBoxLayout,
                               QWidget as _QWidget, QLayout as _QLayout, QApplication as _QApplication)
from PySide6.QtCore import Qt as _Qt, QTimer as _QTimer, QObject as _QObject, Signal as _Signal, QUrl
from PySide6.QtGui import QPalette, QDesktopServices
from argparse import Namespace as _Ns
import sys
import os

from ._default_apps import DefaultThemedApp as _DefaultThemedApp
from .qts import assign_object_names_iterative, AbstractMainWindow, AppStyle, Style, Theme
from .io import SystemTheme, get_system, IOManager

# Standard typing imports for aps
import collections.abc as _a
import typing as _ty
import types as _ts

__all__ = ["QQuickMessageBox", "QBoxDirection", "QNoSpacingBoxLayout", "QtTimidTimer", "BasicAppGUIQt",
           "DefaultAppGUIQt", "QtAppSettings"]

MBoxIcon = _QMessageBox.Icon
MBoxButton = _QMessageBox.StandardButton

class QQuickMessageBox(_QMessageBox):
    """TBA"""
    def __init__(self, parent=None, icon: MBoxIcon | None = None, window_title: str = "", text: str = "",
                 detailed_text: str = "", checkbox: _QCheckBox | None = None,
                 standard_buttons: MBoxButton | None = MBoxButton.Ok,
                 default_button: MBoxButton | None = None):
        """
        An advanced QMessageBox with additional configuration options.

        :param parent: The parent widget.
        :param icon: The icon to display.
        :param window_title: The title of the message box window.
        :param text: The text to display.
        :param detailed_text: The detailed text to display.
        :param checkbox: A QCheckBox instance.
        :param standard_buttons: The standard buttons to include.
        :param default_button: The default button.
        """
        super().__init__(parent=parent)
        for arg, func in zip([standard_buttons, icon, window_title, text, detailed_text, checkbox, default_button],
                             ["setStandardButtons", "setIcon", "setWindowTitle", "setText", "setDetailedText",
                              "setCheckBox", "setDefaultButton"]):
            if arg:
                getattr(self, func)(arg)

        # Set the window to stay on top initially
        self.setWindowState(self.windowState() & ~_Qt.WindowState.WindowMaximized)

        self.raise_()
        self.activateWindow()


QBoxDirection = _QBoxLayout.Direction


class QQuickBoxLayout(_QBoxLayout):
    """TBA"""
    def __init__(self, direction: _QBoxLayout.Direction, spacing: int = 9,
                 margins: tuple[int, int, int, int] = (9, 9, 9, 9), *contents: _QLayout | _QWidget,
                 apply_layout_to: _QWidget | None = None, parent: _QWidget | None = None):
        super().__init__(direction, parent)
        self.setContentsMargins(*margins)
        self.setSpacing(spacing)

        for content in contents:
            if isinstance(content, _QLayout):
                self.addLayout(content)
            elif isinstance(content, _QWidget):
                self.addWidget(content)

        if apply_layout_to is not None:
            apply_layout_to.setLayout(self)


class QNoSpacingBoxLayout(QQuickBoxLayout):
    """TBA"""
    def __init__(self, direction: _QBoxLayout.Direction, *contents: _QLayout | _QWidget,
                 apply_layout_to: _QWidget | None = None, parent: _QWidget | None = None):
        super().__init__(direction, 0, (0, 0, 0, 0), *contents,
                         apply_layout_to=apply_layout_to, parent=parent)


class QtTimidTimer(_QObject):
    """
    A Qt-based version of TimidTimer using QTimer for seamless integration with the Qt event loop.
    """
    timeout = _Signal(int)  # Signal emitted on timeout, with index of timer.

    def __init__(self, parent: _QWidget | None = None) -> None:
        super().__init__(parent)
        self._timers: dict[int, _QTimer | None] = {}  # dict of active timers

    def start(self, interval_ms: int, index: int | None = None):
        """
        Start a timer with the given interval.

        Args:
            interval_ms (int): Timer interval in milliseconds.
            index (int): Optional index to identify the timer.
        """
        timer = _QTimer(self)
        timer.setInterval(interval_ms)
        timer.setSingleShot(False)
        timer.timeout.connect(lambda: self._on_timeout(index if index is not None else len(self._timers)))
        self._timers[index] = timer
        timer.start()

    def stop(self, index: int):
        """
        Stop a timer by its index.

        Args:
            index (int): The index of the timer to stop.
        """
        if index in self._timers:
            self._timers[index].stop()
            self._timers[index].deleteLater()
            self._timers[index] = None
        else:
            raise Exception("Invalid index")

    def _on_timeout(self, index: int):
        """
        Internal method called when a timer times out.

        Args:
            index (int): The index of the timer that timed out.
        """
        self.timeout.emit(index)  # Emit signal with the timer index.

    def stop_all(self):
        """
        Stop all active timers.
        """
        for timer in self._timers.values():
            if timer is not None:
                timer.stop()
                timer.deleteLater()
        self._timers.clear()

    def is_active(self, index: int) -> bool:
        """
        Check if a specific timer is active.

        Args:
            index (int): The index of the timer.

        Returns:
            bool: True if the timer is active, False otherwise.
        """
        return index in self._timers and self._timers[index] is not None and self._timers[index].isActive()


class QtAppSettings(_QObject):
    _instance: _ty.Self | None = None
    _initialized: bool = False
    setup: bool = False

    # Settings Signals
    test_setting_changed = _Signal(str)
    # general
    app_language_changed = _Signal(str)
    window_geometry_changed = _Signal(tuple)  # tuple[int, int, int, int]
    save_window_dimensions_changed = _Signal(bool)
    save_window_position_changed = _Signal(bool)
    # design
    light_theming_changed = _Signal(str)
    dark_theming_changed = _Signal(str)
    font_changed = _Signal(str)
    # advanced
    logging_mode_changed = _Signal(str)

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(QtAppSettings, cls).__new__(cls)
            # cls._instance._initialized = False  # Track initialization state
        return cls._instance

    def __init__(self) -> None:
        if self._initialized:  # Prevent reinitialization
            return
        super().__init__()
        self._initialized = True

    def _initialize(self) -> None:
        raise NotImplementedError()

    def init(self, *args, **kwargs) -> None:
        """Initializes the AppSettings"""
        if self.setup:  # Prevent reinitialization
            return
        self._initialize(*args, **kwargs)
        self._set_default_settings()
        self.setup = True

    def _set_default_settings(self) -> None:
        raise NotImplementedError()

    def _retrieve(self, category: str, name: str) -> _ty.Any:
        raise NotImplementedError()
    def _store(self, category: str, name: str, value: _ty.Any) -> None:
        raise NotImplementedError()

    # Settings Methods
    def get_test_setting(self) -> str:
        return self._retrieve("test", "test_setting")
    def set_test_setting(self, test_setting: str) -> None:
        self._store("test", "test_setting", test_setting)
        self.test_setting_changed.emit(test_setting)
    # general
    def get_app_language(self) -> str:
        return self._retrieve("general", "app_language")
    def set_app_language(self, app_language: str) -> None:
        self._store("general", "app_language", app_language)
        self.app_language_changed.emit(app_language)
    # automatic
    def get_window_geometry(self) -> tuple[int, int, int, int]:
        return self._retrieve("automatic", "window_geometry")  # type: ignore
    def set_window_geometry(self, window_geometry: tuple[int, int, int, int]) -> None:
        self._store("automatic", "window_geometry", window_geometry)
        self.window_geometry_changed.emit(window_geometry)
    # design
    def get_theming(self, mode: SystemTheme) -> str:
        theming_type: str = {SystemTheme.LIGHT: "light_theming",
                             SystemTheme.DARK: "dark_theming"}[mode]
        return self._retrieve("design", theming_type)
    def set_theming(self, mode: SystemTheme, theming: str) -> None:
        theming_type: str = {SystemTheme.LIGHT: "light_theming",
                             SystemTheme.DARK: "dark_theming"}[mode]
        self._store("design", theming_type, theming)
        getattr(self, f"{theming_type}_changed").emit(theming)
    def get_font(self) -> str:
        return self._retrieve("design", "font")
    def set_font(self, font: str) -> None:
        self._store("design", "font", font)
        self.font_changed.emit(font)
    # advanced
    def get_save_window_dimensions(self) -> bool:
        return self._retrieve("advanced", "save_window_dimensions")  # type: ignore
    def set_save_window_dimensions(self, flag: bool) -> None:
        self._store("advanced", "save_window_dimensions", flag)
        self.save_window_dimensions_changed.emit(flag)
    def get_save_window_position(self) -> bool:
        return self._retrieve("advanced", "save_window_position")  # type: ignore
    def set_save_window_position(self, flag: bool) -> None:
        self._store("advanced", "save_window_position", flag)
        self.save_window_position_changed.emit(flag)
    def get_logging_mode(self) -> str:
        return self._retrieve("advanced", "logging_mode")
    def set_logging_mode(self, logging_mode: str) -> None:
        self._settings.store("advanced", "logging_mode", logging_mode)
        self.logging_mode_changed.emit(logging_mode)


class BasicAppGUIQt(_DefaultThemedApp):
    def __init__(self, logs_directory: str, parsed_args: _Ns, logging_level: int, /, setup_thread_pool: bool = False) -> None:
        super().__init__(logs_directory, parsed_args, logging_level, setup_thread_pool=setup_thread_pool)
        try:
            self.qapp: _QtWidgets.QApplication = _QtWidgets.QApplication(sys.argv)  # Just creating the Qapp so we can init widgets
            self.parent = None
            self.timer_number: int = 1
            self.timer: QtTimidTimer = QtTimidTimer()
            self.timer.timeout.connect(self.timer_tick)
        except Exception as e:
            raise Exception("Exception occurred during initialization of the Main class") from e

    def open_url(self, url: str) -> None:
        QDesktopServices.openUrl(QUrl(url))

    def prompt_user(self, title: str, text: str, description: str,
                    level: _ty.Literal["debug", "information", "question", "warning", "error"],
                    options: list[str], default_option: str, checkbox: str | None = None) -> tuple[str | None, bool]:
        if checkbox is not None:
            checkbox = _QtWidgets.QCheckBox(checkbox)
        icon_str = {"information": "Information", "error": "Critical", "question": "Question", "warning": "Warning"}.get(level, "NoIcon")
        icon = getattr(_QMessageBox.Icon, icon_str)

        if self.parent is not None:
            msg_box = QQuickMessageBox(self.parent, icon, title, text,
                                       checkbox=checkbox, standard_buttons=None, default_button=None)
        else:
            msg_box = QQuickMessageBox(None, None, title, text, checkbox=checkbox,
                                       standard_buttons=None, default_button=None)
            window_icon = _QtGui.QIcon(_QtWidgets.QMessageBox.standardIcon(icon))
            msg_box.setWindowIcon(window_icon)

        button_map: dict[str, _QtWidgets.QPushButton] = {}
        for button_str in options:
            button = _QtWidgets.QPushButton(button_str)
            button_map[button_str] = button
            msg_box.addButton(button, _QMessageBox.ButtonRole.ActionRole)
        custom_button = button_map.get(default_option)
        if custom_button is not None:
            msg_box.setDefaultButton(custom_button)
        msg_box.setDetailedText(description)

        clicked_button: int = msg_box.exec()

        checkbox_checked = False
        if checkbox is not None:
            checkbox_checked = checkbox.isChecked()

        for button_text, button_obj in button_map.items():
            if msg_box.clickedButton() == button_obj:
                return button_text, checkbox_checked
        return None, checkbox_checked

    def timer_tick(self, index: int) -> None:
        if index == 0:  # Default 500ms timer
            super().timer_tick()
            self.timer_number += 1
            if self.timer_number > 999:
                self.timer_number = 1

    def exec(self) -> int:
        self.timer.start(500, 0)
        return self.qapp.exec()

    def crash(self, error_title: str, error_text: str, error_description: str) -> bool:
        icon: _QtGui.QIcon
        if error_title == "Warning":
            icon = _QtGui.QIcon(_QtWidgets.QMessageBox.standardIcon(_QtWidgets.QMessageBox.Icon.Warning))
        else:
            icon = _QtGui.QIcon(_QtWidgets.QMessageBox.standardIcon(_QtWidgets.QMessageBox.Icon.Critical))
        custom_icon: bool = False
        if hasattr(self, "abs_window_icon_path"):
            icon_path: str = self.abs_window_icon_path
            icon = _QtGui.QIcon(icon_path)
            custom_icon = True
        msg_box = QQuickMessageBox(None, _QtWidgets.QMessageBox.Icon.Warning if custom_icon else None, error_title, error_text,
                                   error_description,
                                   standard_buttons=_QtWidgets.QMessageBox.StandardButton.Ok | _QtWidgets.QMessageBox.StandardButton.Retry,
                                   default_button=_QtWidgets.QMessageBox.StandardButton.Ok)
        msg_box.setWindowIcon(icon)
        pressed_button = msg_box.exec()
        if pressed_button == _QtWidgets.QMessageBox.StandardButton.Retry:
            return True
        return False

    def close(self) -> None:
        """Cleans up resources"""
        super().close()
        if hasattr(self, "timer"):
            self.timer.stop_all()
        if hasattr(self, "qapp"):
            if self.qapp is not None:
                instance = self.qapp.instance()
                if instance is not None:
                    instance.quit()


class DefaultAppGUIQt(BasicAppGUIQt):
    def __init__(self, themes_directory: str, styles_directory: str, logs_directory: str,
                 parsed_args: _Ns, logging_level: int, /, setup_thread_pool: bool = False, setup_theming: bool = True) -> None:
        super().__init__(logs_directory, parsed_args, logging_level, setup_thread_pool=setup_thread_pool)
        try:
            self.window: AbstractMainWindow
            self.settings: QtAppSettings

            # Setup window
            # self.system: BaseSystemType = get_system()
            # self.os_theme: SystemTheme = self.get_os_theme()
            self.current_theming: str = ""

            self.themes_directory: str = themes_directory
            self.styles_directory: str = styles_directory
            if setup_theming:
                self.load_themes(self.themes_directory)
                self.load_styles(self.styles_directory)
        except Exception as e:
            raise Exception("Exception occurred during initialization of the Main class") from e

    def load_themes(self, theme_folder: str, clear: bool = False) -> None:
        """Loads all theme files from styling/themes"""
        if clear:
            Theme.clear_loaded_themes()
        for file in os.listdir(theme_folder):
            if file.endswith(".qth"):
                path = os.path.join(theme_folder, file)
                Theme.load_from_file(path)

        if Theme.get_loaded_theme("adalfarus::base") is None:
            raise RuntimeError(f"Base theme is not present")

    def load_styles(self, style_folder: str, clear: bool = False) -> None:
        """Loads all styles from styling/styles"""
        if clear:
            Style.clear_loaded_styles()
        for file in os.listdir(style_folder):
            if file.endswith(".qst"):
                path = os.path.join(style_folder, file)
                Style.load_from_file(path)

        if (Style.get_loaded_style("Default Dark", "*") is None
                or Style.get_loaded_style("Default Light", "*") is None):
            raise RuntimeError(f"Default light and/or dark style are/is not present")

    def apply_theme(self) -> None:
        theming_str: str = self.settings.get_theming(self.os_theme)
        self.current_theming = theming_str
        theme_str, style_str = theming_str.split("/", maxsplit=1)
        theme: Theme | None = Theme.get_loaded_theme(theme_str)

        if theme is None:  # TODO: Popup
            IOManager().warning(f"Specified theme '{theme}' is not available", "", show_dialog=True)
            return
        style: Style | None = theme.get_compatible_style(style_str.replace("_", " ").title())
        if style is None:
            IOManager().warning(f"Couldn't find specified style {style_str} for theme {theme_str}", "",
                                show_dialog=True)
            return
        theme_str, palette = theme.apply_style(style, QPalette(),
                                               transparency_mode="none")  # TODO: Get from settings
        self.qapp.setPalette(palette)
        self.window.set_global_theme(theme_str, getattr(self.window.AppStyle, theme.get_base_styling()))

    def check_theme_change(self):
        if self.timer_number & 1 == 1:
            current_os_theme = self.get_os_theme()
            if current_os_theme != self.os_theme or self.settings.get_theming(current_os_theme) != self.current_theming:
                self.os_theme = current_os_theme
                self.apply_theme()

    def timer_tick(self, index: int) -> None:
        super().timer_tick(index)
        if index == 0:  # Default 500ms timer
            self.check_theme_change()

    def exec(self) -> int:
        self.parent = self.window.internal_obj()

        self.window.setup_gui()

        x, y, width, height = self.settings.get_window_geometry()
        if not self.settings.get_save_window_dimensions():
            width = 1050
            height = 640
        if self.settings.get_save_window_position():
            self.window.set_window_geometry(x, y + 31, width, height)  # Somehow saves it as 31 pixels less,
        else:  # I guess windows does some weird shit with the title bar
            self.window.set_window_dimensions(width, height)

        assign_object_names_iterative(self.window.internal_obj())  # Set object names for theming

        self.window.app = self.qapp
        self.apply_theme()
        self.window.start()  # Shows gui
        self.window.set_font(self.settings.get_font())  # So that everything gets updated
        return super().exec()



"""Web links to aplustools"""
from aplustools.web.request import fetch, Result, BatchRequestHandler
from aplustools.web.utils import WebPage, url_validator

__all__ = ["fetch", "Result", "BatchRequestHandler", "WebPage", "url_validator"]



"""IO Component of dancer"""
from pathlib import Path as PLPath
import logging
import sys
import re
import io
import os

from aplustools.data import SingletonMeta
from aplustools.io.env import BaseSystemType, SystemTheme, get_system

# Standard typing imports for aps
import collections.abc as _a
import typing as _ty
import types as _ts

__all__ = ["BaseSystemType", "SystemTheme", "get_system", "SingletonMeta", "ActLogger", "IOManager"]


# Copyright adalfarus
# Helper class to redirect streams to the logger
class _StreamToLogger(io.IOBase):
    """
    File-like object that redirects writes to a logger instance.
    """

    def __init__(self, logger, log_level, original_stream: _ty.IO):
        """
        Initialize the stream redirection.

        :param logger: Logger instance where messages will be redirected.
        :param log_level: Logging level for the redirected messages.
        """
        self.logger = logger
        self.log_level = log_level
        self.original_stream = original_stream
        self.linebuf = ""

    def write(self, buf):
        """
        Write method for the file-like object.

        :param buf: String to write.
        """
        self.linebuf += buf
        while "\n" in self.linebuf:
            line, self.linebuf = self.linebuf.split("\n", 1)
            if line:
                self.logger.log(self.log_level, line)
            else:
                # Handle empty lines (e.g., when there are multiple newlines)
                self.logger.log(self.log_level, "")

    def flush(self):
        """
        Flush method for file-like object.
        """
        if self.linebuf:
            self.logger.log(self.log_level, self.linebuf.rstrip())
            self.linebuf = ""

    def restore(self) -> io.IOBase:
        return self.original_stream

# Copyright adalfarus
class ActLogger(metaclass=SingletonMeta):
    """
    A configurable logger for ActFramework that supports logging to both the console
    and an optional log file. It provides methods to log messages at different
    logging levels and can monitor and redirect system output streams.

    Attributes:
        _logger: The logger instance used for logging messages.
        handlers: A list of handlers attached to the logger (console, file handlers).
    """
    def __init__(self, name: str = "ActLogger", log_to_file: bool = False, filepath: str | PLPath = "app.log") -> None:
        """
        Initialize the act logger.

        :param name: Name of the logger.
        :param log_to_file: Boolean indicating if logs should be written to a file.
        :param filepath: Path of the log file.
        """
        self._logger = logging.getLogger(name)
        self._logger.setLevel(logging.DEBUG)
        self.handlers: list[logging.Handler] = []

        # Create formatter with the desired format
        formatter = logging.Formatter(
            '[%(asctime)s.%(msecs)03d] [%(levelname)s] %(message)s',
            datefmt='%Y-%m-%d %H:%M:%S'
        )

        # Console handler
        console_handler = logging.StreamHandler(sys.__stdout__)
        console_handler.setFormatter(formatter)
        self._logger.addHandler(console_handler)
        self.handlers.append(console_handler)

        # File handler (optional)
        if log_to_file:
            file_handler = logging.FileHandler(filepath, encoding='utf8')
            file_handler.setFormatter(formatter)
            self._logger.addHandler(file_handler)
            self.handlers.append(file_handler)
        self.logging_level: int = -1

    def create_pipe_redirect(self, pipe: _ty.IO, level: int = logging.INFO) -> _StreamToLogger:
        """
        Return a stream wrapper that redirects writes to the logger.

        :param pipe: The original stream (e.g., sys.stdout, a file, etc.)
        :param level: Logging level to use
        :return: _StreamToLogger instance
        """
        return _StreamToLogger(self._logger, level, pipe)

    def restore_pipe(self, replacement: _StreamToLogger) -> io.IOBase:
        """
        Restore the given global pipe (e.g. sys.stdout) to its original value.

        :param replacement: The _StreamToLogger that was used to override it.
        """
        return replacement.restore()

    def add_handler(self, mirror_to_io: logging.Handler) -> None:
        """Adds a handler to the underlying logger and keeps track of it in the .handlers attribute"""
        self.handlers.append(mirror_to_io)
        self._logger.addHandler(mirror_to_io)

    def log(self, level: int, message: str) -> None:
        """Log a message with a specific logging level."""
        self._logger.log(level, message)

    # Convenience methods for different logging levels
    def info(self, message: str) -> None:
        """
        Log an informational message.

        :param message: The message to log at the INFO level. Typically used for
                        general application progress or operational messages.
        """
        self._logger.info(message)

    def debug(self, message: str) -> None:
        """
        Log a debug message.

        :param message: The message to log at the DEBUG level. Typically used for
                        detailed information useful for diagnosing issues or tracing execution flow.
        """
        self._logger.debug(message)

    def error(self, message: str) -> None:
        """
        Log an error message.

        :param message: The message to log at the ERROR level. Typically used to
                        indicate a significant issue or error in the application.
        """
        self._logger.error(message)

    def warning(self, message: str) -> None:
        """
        Log a warning message.

        :param message: The message to log at the WARNING level. Typically used to
                        indicate a slight issue or warn about an action.
        :return:
        """
        self._logger.warning(message)

    def setLevel(self, logging_level: int) -> None:
        """Sets the level of the underlying logger"""
        self._logger.setLevel(logging_level)
        self.logging_level = logging_level

# Copyright zScout
T = _ty.TypeVar("T")
class OrderedSet(_ty.Generic[T]):
    # Docs generated with Github Copilot
    def __init__(self, iterable: _ty.Iterable[_ty.Any] | None = None) -> None:
        """
        OrderedSet is a hybrid of list and set. It maintains the order of elements like a list and ensures that each
        element is unique like a set. It is implemented using a list and a set. The list maintains the order of elements
        and the set ensures that each element is unique. The list is used to maintain the order of elements and the set is
        used to check if an element is already present in the OrderedSet."""
        self._items: list[T] = []
        self._seen: set[T] = set()
        if iterable:
            for item in iterable:
                self.add(item)

    def add(self, item: T) -> None:
        """
        Adds an item to the OrderedSet if it is not already present in the OrderedSet.
        :param item: The item to be added to the OrderedSet.
        :return: None
        """
        if item not in self._seen:
            self._items.append(item)
            self._seen.add(item)

    def discard(self, item: T) -> None:
        """
        Removes an item from the OrderedSet if it is present in the OrderedSet.
        :param item: The item to be removed from the OrderedSet.
        :return: None
        """
        if item in self._seen:
            self._items.remove(item)
            self._seen.remove(item)

    def remove(self, item: T) -> None:
        """
        Removes an item from the OrderedSet if it is present in the OrderedSet. If the item is not present in the
        OrderedSet, a KeyError is raised.
        :param item: The item to be removed from the OrderedSet.
        :return: None
        """
        if item not in self._seen:
            raise KeyError(f"{item} not in OrderedSet")
        self.discard(item)

    def clear(self) -> None:
        """
        Removes all items from the OrderedSet.
        :return: None
        """
        self._items.clear()
        self._seen.clear()

    def get_index(self, item: T) -> int:
        """
        Returns the index of an item in the OrderedSet.
        :param item: The item whose index is to be returned.
        :return: The index of the item in the OrderedSet.
        """
        return self._items.index(item)

    def get_by_index(self, index: int) -> T:
        """
        Returns the item at a given index in the OrderedSet.
        :param index: The index of the item to be returned.
        :return: The item at the given index in the OrderedSet.
        """
        return self._items[index]

    def to_list(self) -> list[T]:
        """
        Returns the items in the OrderedSet as a list.
        :return: The items in the OrderedSet as a list.
        """
        return self._items

    def to_set(self) -> set[T]:
        """
        Returns the items in the OrderedSet as a set.
        :return: The items in the OrderedSet as a set.
        """
        return self._seen

    @staticmethod
    def from_list(lst: _ty.List[T]) -> 'OrderedSet':
        return OrderedSet(lst)

    @staticmethod
    def from_set(st: _ty.Set[T]) -> 'OrderedSet':
        return OrderedSet(st)

    def __len__(self):
        """
        Returns the number of items in the OrderedSet.
        :return: The number of items in the OrderedSet.
        """
        return len(self._items)

    def __iter__(self):
        """
        Returns an iterator over the items in the OrderedSet.
        :return: An iterator over the items in the OrderedSet.
        """
        return iter(self._items)

    def __contains__(self, item: T):
        """
        Returns True if an item is present in the OrderedSet, False otherwise.
        :param item: The item to be checked for presence in the OrderedSet.
        :return: True if the item is present in the OrderedSet, False otherwise.
        """
        return item in self._seen

    def __repr__(self):
        """
        Returns a string representation of the OrderedSet.
        :return: A string representation of the OrderedSet.
        """
        return f"OrderedSet({self._items})"

    def __eq__(self, other):
        """
        Returns True if the OrderedSet is equal to another OrderedSet, False otherwise.
        :param other: The other OrderedSet to be compared with.
        :return: True if the OrderedSet is equal to the other OrderedSet, False otherwise.
        """
        if isinstance(other, OrderedSet):
            return self._items == other._items
        return False

    def __or__(self, other):
        """
        Returns the union of the OrderedSet with another OrderedSet or set.
        :param other: The other OrderedSet or set to be unioned with.
        :return: The union of the OrderedSet with the other OrderedSet or set.
        """
        if not isinstance(other, (OrderedSet, set)):
            return NotImplemented
        return OrderedSet(self._items + [item for item in other if item not in self])

    def __and__(self, other):
        """
        Returns the intersection of the OrderedSet with another OrderedSet or set.
        :param other: The other OrderedSet or set to be intersected with.
        :return: The intersection of the OrderedSet with the other OrderedSet or set.
        """
        if not isinstance(other, (OrderedSet, set)):
            return NotImplemented
        return OrderedSet(item for item in self if item in other)

    def __sub__(self, other):
        """
        Returns the difference of the OrderedSet with another OrderedSet or set.
        :param other: The other OrderedSet or set to be differenced with.
        :return: The difference of the OrderedSet with the other OrderedSet or set.
        """
        if not isinstance(other, (OrderedSet, set)):
            return NotImplemented
        return OrderedSet(item for item in self if item not in other)

# Copyright zScout
S = _ty.TypeVar("S")
class StaticContainer(_ty.Generic[S]):
    def __init__(self, value: S | None = None) -> None:
        self._value: S | None = value

    # TODO: self._value is S or None but returned it is just S?
    def get_value(self) -> S:
        """
        Get the value stored in the Container
        :return: (S) returns the current value stored
        """
        return self._value

    def set_value(self, new_value: S) -> None:
        """
        Sets a new value to store in the Container
        :param new_value: S
        :return: None
        """
        self._value = new_value

    def has_value(self) -> bool:
        """
        Returns a bool to indicate if the container stores a value
        :return: Srue, if the Container stores a value
        """
        return self._value is not None

    def clear_value(self) -> None:
        """
        Clears the value stored
        :return: None
        """
        self._value = None

# Copyright zScout  TODO: Refactor, title etc are just wrongly ordered in the methods and adding support for custom icons; Or also ignoring msgs with only small changes?
class IOManager(metaclass=SingletonMeta):
    """TBA"""
    _do_not_show_again: OrderedSet[str] = OrderedSet()
    _currently_displayed: OrderedSet[str] = OrderedSet()
    _button_display_callable: StaticContainer[_ty.Callable] = StaticContainer()
    _is_indev: StaticContainer[bool] = StaticContainer()
    _popup_queue: _ty.List[_ty.Callable[[_ty.Any], _ty.Any]] = []

    _logger: ActLogger

    def add_handler(self, handler) -> None:
        self._logger.add_handler(handler)

    def has_cached_errors(self) -> bool:
        """
        Returns if there are popups cached which have not been displayed yet
        :return: bool
        """
        return len(self._popup_queue) > 0

    def invoke_prompts(self) -> None:
        """

        :return:
        """
        if not self.has_cached_errors():
            return

        popup_callable: _ty.Callable = self._popup_queue[0]
        self._popup_queue.pop(0)

        popup_callable()

    def init(self, promt_creation_callable: _ty.Callable, logs_folder_path: str, is_indev: bool) -> None:
        """
        Initializes the ErrorCache with a popup creation callable and development mode flag.
        :param promt_creation_callable: Callable used to create popups.
        :param logs_folder_path: File path to the logs folder.
        :param is_indev: Boolean indicating whether the application is in development mode.
        :return: None
        """
        self._button_display_callable.set_value(promt_creation_callable)
        self._order_logs(logs_folder_path)
        self._logger = ActLogger(log_to_file=True, filepath=os.path.join(logs_folder_path, "latest.log"))
        sys.stdout = self._logger.create_pipe_redirect(sys.stdout, level=logging.DEBUG)
        sys.stderr = self._logger.create_pipe_redirect(sys.stderr, level=logging.ERROR)
        # Replace fancy characters
        self._is_indev.set_value(is_indev)

    def set_logging_level(self, level: int) -> None:
        """
        Sets the logging level of the Logger
        :param level: Logging level to set to.
        :return: None
        """
        self._logger.setLevel(level)

    def get_logging_level(self) -> int:
        return self._logger.logging_level

    @staticmethod
    def _order_logs(directory: str) -> None:
        logs_dir = PLPath(directory)
        to_log_file = logs_dir / "latest.log"

        if not to_log_file.exists():
            print("Logfile missing")
            return

        with open(to_log_file, "rb") as f:
            # (solution from https://stackoverflow.com/questions/46258499/how-to-read-the-last-line-of-a-file-in-python)
            first_line = f.readline().decode()
            try:  # catch OSError in case of a one line file
                f.seek(-2, os.SEEK_END)
                while f.read(1) != b"\n":
                    f.seek(-2, os.SEEK_CUR)
            except OSError:
                f.seek(0)
            last_line = f.readline().decode()

        try:
            date_pattern = r"^[\[(](\d{4}-\d{2}-\d{2})"
            start_date = re.search(date_pattern, first_line).group(1)  # type: ignore
            end_date = re.search(date_pattern, last_line).group(1)  # type: ignore
        except AttributeError:
            print("Removing malformed latest.log")
            os.remove(to_log_file)
            return

        date_name = f"{start_date}_{end_date}"
        date_logs = list(logs_dir.rglob(f"{date_name}*.log"))

        if not date_logs:
            new_log_file_name = logs_dir / f"{date_name}.log"
        else:
            try:
                max_num = max(
                    (int(re.search(r"#(\d+)$", p.stem).group(1)) for p in date_logs if  # type: ignore
                     re.search(r"#(\d+)$", p.stem)),
                    default=0
                )
            except AttributeError:
                print("AttribError")
                return
            max_num += 1
            base_log_file = logs_dir / f"{date_name}.log"
            if base_log_file.exists():
                os.rename(base_log_file, logs_dir / f"{date_name}#{max_num}.log")
                max_num += 1
            new_log_file_name = logs_dir / f"{date_name}#{max_num}.log"

        os.rename(to_log_file, new_log_file_name)
        print(f"Renamed latest.log to {new_log_file_name}")

    def _show_prompt(self, title: str, text: str, description: str,
                     level: _ty.Literal["debug", "information", "question", "warning", "error"],
                     custom_options: _ty.Dict[str, _ty.Callable] | None = None) -> None:
        """
        Displays a dialog box with the provided information.
        :param title: Title of the dialog box.
        :param text: Main text content of the dialog box.
        :param description: Additional description text.
        :param level: Type of icon to display in the dialog box.
        :return: None
        """
        if text in self._currently_displayed:
            # Error is currently displayed
            return

        if text in self._do_not_show_again:
            # Error should not be displayed again
            return

        if not self._button_display_callable.has_value():
            return

        self._currently_displayed.add(text)

        checkbox_text: str = "Do not show again"
        options_list: _ty.List[str] = ["Ok"]
        default_option: str = options_list[0]

        # add custom buttons
        if custom_options is not None:
            for key in list(custom_options.keys()):
                options_list.append(key)

        popup_creation_callable: _ty.Callable = self._button_display_callable.get_value()
        popup_return: tuple[str | None, bool] = popup_creation_callable(title, text, description, level,  # "[N.E.F.S] " +
                                                                        options_list, default_option, checkbox_text)

        if popup_return[1]:
            self._do_not_show_again.add(text)
        self._currently_displayed.remove(text)

        # invoke button commands
        option_name: str = popup_return[0]
        if custom_options is not None:
            if option_name in custom_options:
                custom_options[option_name]()

    def _handle_prompt(self, show_prompt: bool, title: str, log_message: str, description: str,
                       level: _ty.Literal["debug", "information", "question", "warning", "error"],
                       custom_options: _ty.Dict[str, _ty.Callable] | None = None) -> None:
        """
        Handles the process of displaying a dialog based on parameters.
        :param show_prompt: Boolean indicating whether to show the dialog.
        :param title: Title of the dialog.
        :param log_message: Log message associated with the dialog.
        :param description: Additional description text.
        :param level: Type of icon to display in the dialog.
        :return: None
        """
        if not show_prompt:
            return

        self._popup_queue.append(lambda: self._show_prompt(title, log_message, description, level, custom_options))

    # "Errors"

    def warn(self, log_message: str, description: str = "", show_prompt: bool = False,
             print_log: bool = True, prompt_title: str | None = None,
             custom_options: _ty.Dict[str, _ty.Callable] | None = None) -> None:
        """
        Logs a warning message and optionally displays a warning dialog.
        :param prompt_title: Sets the popup window title
        :param custom_options: Defines additional buttons for the popup window
        :param log_message: The warning message to log.
        :param description: Additional description of the warning.
        :param show_prompt: Whether to show a dialog for the warning.
        :param print_log: Whether to print the log message.
        :return: None
        """
        return self.warning(log_message, description, show_prompt, print_log, prompt_title, custom_options)

    def info(self, log_message: str, description: str = "", show_prompt: bool = False,
             print_log: bool = True, prompt_title: str | None = None,
             custom_options: _ty.Dict[str, _ty.Callable] | None = None) -> None:
        """
        Logs an informational message and optionally displays an information dialog.
        :param log_message: The informational message to log.
        :param description: Additional description of the information.
        :param show_prompt: Whether to show a dialog for the information.
        :param print_log: Whether to print the log message.
        :param prompt_title: Sets the popup window title
        :param custom_options: Defines additional buttons for the popup window
        :return: None
        """
        title: str = "Information"
        if prompt_title is not None:
            title += f": {prompt_title}"

        if print_log:
            self._logger.info(f"{log_message} {f'({description})' if description else ''}")

        if ActLogger().logging_level > logging.INFO:
            return

        self._handle_prompt(show_prompt, title, log_message, description, "information", custom_options)

    def warning(self, log_message: str, description: str = "", show_prompt: bool = False,
                print_log: bool = True, prompt_title: str | None = None,
                custom_options: _ty.Dict[str, _ty.Callable] | None = None) -> None:
        """
        Logs a warning message and optionally displays a warning dialog.
        :param log_message: The warning message to log.
        :param description: Additional description of the warning.
        :param show_prompt: Whether to show a dialog for the warning.
        :param print_log: Whether to print the log message.
        :param prompt_title: Sets the popup window title
        :param custom_options: Defines additional buttons for the popup window
        :return: None
        """
        title: str = "Warning"
        if prompt_title is not None:
            title += f": {prompt_title}"

        if print_log:
            self._logger.warning(f"{log_message} {f'({description})' if description else ''}")

        if ActLogger().logging_level > logging.WARNING:
            return

        self._handle_prompt(show_prompt, title, log_message, description, "warning", custom_options)

    def fatal_error(self, log_message: str, description: str = "", show_prompt: bool = False,
                    print_log: bool = True, prompt_title: str | None = None,
                    custom_options: _ty.Dict[str, _ty.Callable] | None = None) -> None:
        """
        Logs a fatal error message and optionally displays an error dialog.
        :param log_message: The error message to log.
        :param description: Additional description of the error.
        :param show_prompt: Whether to show a dialog for the error.
        :param print_log: Whether to print the log message.
        :param prompt_title: Sets the popup window title
        :param custom_options: Defines additional buttons for the popup window
        :return: None
        """
        self.error(log_message, description, show_prompt, print_log, prompt_title, custom_options, error_severity="FATAL")

    def error(self, log_message: str, description: str = "", show_prompt: bool = False,
              print_log: bool = True, prompt_title: str | None = None,
              custom_options: _ty.Dict[str, _ty.Callable] | None = None, *_,
              error_severity: str = "NORMAL") -> None:
        """
        Logs an error message and optionally displays an error dialog.
        :param log_message: The error message to log.
        :param description: Additional description of the error.
        :param show_prompt: Whether to show a dialog for the error.
        :param print_log: Whether to print the log message.
        :param prompt_title: Sets the popup window title
        :param custom_options: Defines additional buttons for the popup window
        :param error_severity: Defined a custom error name.
        :return: None
        """
        title: str = f"{str(error_severity).capitalize()} Error"
        if prompt_title is not None:
            title += f": {prompt_title}"

        if print_log:
            self._logger.error(f"{str(error_severity)}: {log_message} {f'({description})' if description else ''}")

        if ActLogger().logging_level > logging.ERROR:
            return

        self._handle_prompt(show_prompt, title, log_message, description, "error", custom_options)

    def debug(self, log_message: str, description: str = "", show_prompt: bool = False,
              print_log: bool = True, prompt_title: str | None = None,
              custom_options: _ty.Dict[str, _ty.Callable] | None = None) -> None:
        """
        Logs a debug message and optionally displays a debug dialog, only if in development mode.
        :param log_message: The debug message to log.
        :param description: Additional description of the debug information.
        :param show_prompt: Whether to show a dialog for the debug information.
        :param print_log: Whether to print the log message.
        :param prompt_title: Sets the popup window title
        :param custom_options: Defines additional buttons for the popup window
        :return: None
        """
        if not self._is_indev.has_value():
            return

        INDEV: bool = self._is_indev.get_value()  # config.INDEV
        if not INDEV:
            return

        title: str = "Debug"
        if prompt_title is not None:
            title += f": {prompt_title}"

        if print_log:
            self._logger.debug(f"{log_message} {f'({description})' if description else ''}")

        if ActLogger().logging_level > logging.DEBUG:
            return

        self._handle_prompt(show_prompt, title, log_message, description, "debug", custom_options)

    def prompt_user(self, title: str, message: str, details: str,
                    level: _ty.Literal["debug", "information", "question", "warning", "error"],
                    options: list[str], default_option: str, checkbox_label: str | None = None,
                    return_type: _ty.Literal["thread", "proc"] = "thread"):  # -> tuple[SafeList, Event]:
        raise NotImplementedError("The current implementation does not support user prompting with arbitrary data.")

    def __del__(self) -> None:
        if hasattr(self, "_logger"):
            sys.stdout = self._logger.restore_pipe(sys.stdout)  # type: ignore
            sys.stderr = self._logger.restore_pipe(sys.stderr)  # type: ignore



"""System links to aplustools"""
from aplustools.io.env import (SystemTheme, BasicSystemFunctions, is_compiled, get_system, diagnose_shutdown_blockers,
                               is_accessible, suppress_warnings, MAX_PATH, BaseSystemType, auto_repr,
                               auto_repr_with_privates)
from aplustools.io.fileio import SafeFileWriter, BasicFileLock, OSFileLock, os_open, BasicFDWrapper

__all__ = ["SystemTheme", "BasicSystemFunctions", "is_compiled", "get_system", "diagnose_shutdown_blockers",
           "is_accessible", "suppress_warnings", "MAX_PATH", "BaseSystemType", "auto_repr", "auto_repr_with_privates",
           "SafeFileWriter", "BasicFileLock", "OSFileLock", "os_open", "BasicFDWrapper"]



"""Configures your environment"""
import enum
import warnings
from dataclasses import dataclass as _dc
import platform
import shutil
import sys
import os
import re

import typing as _ty
import types as _ts
import typing_extensions as _tx

@_dc(frozen=True)
class OSEntry:
    """Contains all info on the OS"""
    major_version: str
    minor_versions: tuple[str, ...]
    machines: tuple[str, ...]

_DirectoryTree = dict[str, _ty.Union["_DirectoryTree", None]]
PlatformVersionsType = dict[str, tuple[tuple[str, ...], tuple[str, ...]]]
OSListType = dict[str, list[OSEntry]]  # dict[str, PlatformVersionsType]

INDEV: bool
INDEV_KEEP_RUNTIME_FILES: bool
COMPILE_TIME_FLAGS: bool
PROGRAM_NAME: str
VERSION: int
VERSION_ADD: str
PROGRAM_NAME_NORMALIZED: str
WORKING_OS_LIST: OSListType
UNTESTED_OS_LIST: OSListType
INCOMPATIBLE_OS_LIST: OSListType
PY_LIST: list[tuple[int, int]]
DIR_STRUCTURE: _DirectoryTree
LOCAL_MODULE_LOCATIONS: list[str]

RUNTIME_FILE_EXTENSIONS: tuple[str, ...] = (
    ".json",
    ".yml", ".yaml",
    ".ini",
    ".toml",
    ".conf",
    ".cfg",
    ".xml",
    ".log",
    ".err",
    ".out",
    ".trace",
    ".dmp",
    ".db",
    ".sqlite", ".sqlite3",
    ".dat",
    ".bak",
    ".tmp",
    ".cache",
    ".mo", ".po",
    ".lang",
    ".key", ".pem",
    ".cred", ".auth",
    ".unicode",
    ".map",
    ".layout",
    ".mod",
    ".profile",
    ".session"
)

OLD_CWD: str = os.getcwd()
if "CONFIG_DONE" not in locals():
    CONFIG_DONE: bool = False
if "CHECK_DONE" not in locals():
    CHECK_DONE: bool = False

exported_logs: str
base_app_dir: str
old_cwd: str

exit_code: int
exit_message: str

def is_compiled() -> bool:
    """  # From aps.io.env
    Detects if the code is running in a compiled environment and identifies the compiler used.

    This function checks for the presence of attributes and environment variables specific
    to common Python compilers, including PyInstaller, cx_Freeze, and py2exe.
    :return: bool
    """
    return getattr(sys, "frozen", False) and (hasattr(sys, "_MEIPASS") or sys.executable.endswith(".exe"))

def get_version_str() -> str:
    """Returns VERSION + VERSION_ADD as a string"""
    return str(VERSION) + VERSION_ADD

# From aplustools.io.env
def _get_appdata_dir(app_dir: str, scope: _ty.Literal["user", "global"] = "global"):
    system = platform.system()
    if system == "Windows":
        if scope == "user":
            return os.path.join(
                os.environ.get("APPDATA"), app_dir
            )  # App data for the current user
        return os.path.join(
            os.environ.get("PROGRAMDATA"), app_dir
        )  # App data for all users
    elif system == "Darwin":
        if scope == "user":
            return os.path.join(
                os.path.expanduser("~"), "Library", "Application Support", app_dir
            )  # App data for the current user
        return os.path.join(
            "/Library/Application Support", app_dir
        )  # App data for all users
    elif system == "Linux":
        if scope == "user":
            return os.path.join(
                os.path.expanduser("~"), ".local", "share", app_dir
            )  # App data for the current user
        return os.path.join("/usr/local/share", app_dir)  # App data for all users
    elif system == "FreeBSD":
        if scope == "user":
            return os.path.join(os.path.expanduser("~"), ".local", "share", app_dir)
        return os.path.join("/usr/local/share", app_dir)

def _configure() -> dict[str, str]:
    if is_compiled():
        os.chdir(os.path.dirname(os.path.abspath(__file__)))
        if not sys.stdout:
            sys.stdout = open(os.devnull, "w")
        if not sys.stderr:
            sys.stderr = open(os.devnull, "w")

    if COMPILE_TIME_FLAGS:
        import multiprocessing
        multiprocessing.freeze_support()

    accumulated_logs = "Starting cloning of defaults ...\n"
    old_cwd = os.getcwd()
    install_dir = os.path.join(old_cwd, "default-config")
    # So we never have a relative dir, even if something fails
    base_app_dir = os.path.abspath(_get_appdata_dir(f"{PROGRAM_NAME_NORMALIZED}_{VERSION}{VERSION_ADD}", "user"))

    if INDEV and os.path.exists(base_app_dir):  # Remove everything to simulate a fresh install
        if not INDEV_KEEP_RUNTIME_FILES:
            shutil.rmtree(base_app_dir)
            os.mkdir(base_app_dir)
        else:  # Skip only .db or .log files
            for root, dirs, files in os.walk(base_app_dir, topdown=False):
                for file in files:
                    if not file.endswith(RUNTIME_FILE_EXTENSIONS):
                        os.remove(os.path.join(root, file))
                for directory in dirs:
                    dir_path = os.path.join(root, directory)
                    if not any(f.endswith(RUNTIME_FILE_EXTENSIONS) or os.path.isdir(os.path.join(dir_path, f)) for f in os.listdir(dir_path)):
                        shutil.rmtree(dir_path)

    dirs_to_create = []
    # Use a stack to iteratively traverse the directory structure
    stack: list[tuple[str, _DirectoryTree]] = [(base_app_dir, DIR_STRUCTURE)]
    while stack:
        current_base, subtree = stack.pop()
        for name, children in subtree.items():
            current_path = os.path.join(current_base, name)
            dirs_to_create.append(current_path)
            accumulated_logs += f"Cloning {current_path}\n"
            if isinstance(children, dict) and children:
                stack.append((current_path, children))
        # base_path, (dir_name, subdirs) = stack.pop()
        # current_path = os.path.join(base_path, dir_name)
        #
        # if not subdirs:  # No subdirectories; it's a leaf
        #     dirs_to_create.append(current_path)
        #     accumulated_logs += f"Cloning {current_path}\n"
        # else:
        #     for subdir in subdirs:  # Add each subdirectory to the stack for further processing
        #         if isinstance(subdir, tuple):
        #             stack.append((current_path, subdir))  # Nested structure
        #         else:  # Direct leaf under the current directory
        #             dirs_to_create.append(os.path.join(current_path, subdir))
        #             accumulated_logs += f"Cloning {os.path.join(current_path, subdir)}\n"
    for dir_to_create in dirs_to_create:
        os.makedirs(dir_to_create, exist_ok=True)
    for loc in LOCAL_MODULE_LOCATIONS:
        sys.path.insert(0, os.path.join(base_app_dir, loc))

    for dirpath, dirnames, filenames in os.walk(install_dir):
        for filename in filenames:
            file_path = os.path.join(dirpath, filename)
            stripped_filename = os.path.relpath(file_path, install_dir)
            alternate_file_location = os.path.join(base_app_dir, stripped_filename)
            if not os.path.exists(alternate_file_location) or INDEV:  # Replace all for indev
                # accumulated_logs += f"{file_path} -> {alternate_file_location}\n"  # To flush config prints in main
                os.makedirs(os.path.dirname(alternate_file_location), exist_ok=True)
                shutil.copyfile(file_path, alternate_file_location)
            # else:
            #     accumulated_logs += f"{alternate_file_location} Already exists\n"  # To flush config prints in main

    os.chdir(base_app_dir)
    return {
        "accumulated_logs": accumulated_logs, "old_cwd": old_cwd, "install_dir": install_dir,
        "base_app_dir": base_app_dir,
    }

@_dc(frozen=True)
class AppConfig:
    """Contains all info on the app"""
    INDEV: bool
    INDEV_KEEP_RUNTIME_FILES: bool
    COMPILE_TIME_FLAGS: bool
    PROGRAM_NAME: str
    PROGRAM_NAME_NORMALIZED: str
    VERSION: int
    VERSION_ADD: str
    WORKING_OS_LIST: OSListType
    UNTESTED_OS_LIST: OSListType
    INCOMPATIBLE_OS_LIST: OSListType
    PY_LIST: list[tuple[int, int]]
    DIR_STRUCTURE: _DirectoryTree
    LOCAL_MODULE_LOCATIONS: list[str]

def configure(app_config: AppConfig) -> None:
    """Configures all the information into the config module"""
    global INDEV, INDEV_KEEP_RUNTIME_FILES, COMPILE_TIME_FLAGS, PROGRAM_NAME, VERSION, VERSION_ADD, \
        PROGRAM_NAME_NORMALIZED,  WORKING_OS_LIST, UNTESTED_OS_LIST, INCOMPATIBLE_OS_LIST, PY_LIST, DIR_STRUCTURE, \
        LOCAL_MODULE_LOCATIONS
    INDEV = app_config.INDEV
    INDEV_KEEP_RUNTIME_FILES = app_config.INDEV_KEEP_RUNTIME_FILES
    COMPILE_TIME_FLAGS = app_config.COMPILE_TIME_FLAGS
    PROGRAM_NAME = app_config.PROGRAM_NAME
    VERSION = app_config.VERSION
    VERSION_ADD = app_config.VERSION_ADD
    PROGRAM_NAME_NORMALIZED = app_config.PROGRAM_NAME_NORMALIZED
    WORKING_OS_LIST = app_config.WORKING_OS_LIST
    UNTESTED_OS_LIST = app_config.UNTESTED_OS_LIST
    INCOMPATIBLE_OS_LIST = app_config.INCOMPATIBLE_OS_LIST
    PY_LIST = app_config.PY_LIST
    DIR_STRUCTURE = app_config.DIR_STRUCTURE
    LOCAL_MODULE_LOCATIONS = app_config.LOCAL_MODULE_LOCATIONS

class OSListExitState(enum.Enum):
    SystemNotSupported = 0
    ReleaseNotSupported = 1
    VersionNotSupported = 2
    ReleaseOrVersionNotSupported = 3
    Supported = 4

#T = _ty.TypeVar("T")
#def get_from_os_list(os_list: OSListType, key: str, base: T = None) -> OSEntry | T:
#    for osl in os_list:
#        if osl.os_name == key:
#            return osl
#    return base

def _check_os_list(system: str, release: str, version: str, machine: str, os_list: OSListType) -> tuple[OSListExitState, str, str, str]:
    platform_versions: list[OSEntry] | None = os_list.get(system, None)

    if platform_versions is None:
        return OSListExitState.SystemNotSupported, "", "", ""

    for ose in platform_versions:
        possible_major, (possible_minors, possible_machines) = ose.major_version, (ose.minor_versions, ose.machines)
        if re.fullmatch(possible_major, release):
            minor_matches = [re.fullmatch(possible_minor, version) is not None for possible_minor in
                       possible_minors]

            if any(minor_matches) or possible_minors == ("any",):
                machine_matches = [re.fullmatch(possible_machine, machine) is not None for possible_machine in
                           possible_machines]

                if any(machine_matches) or possible_machines == ("any",):
                    return OSListExitState.Supported, release, version, machine  # Currently, this is redundant, maybe not in the future

    return OSListExitState.ReleaseOrVersionNotSupported, "", "", ""

def _format_os_list(os_list: OSListType) -> str:
    return ", ".join(os_list.keys())

def check() -> RuntimeError | UserWarning | str:
    """Check if environment is suitable"""
    global CHECK_DONE, exit_code, exit_message

    if CHECK_DONE:
        return "None"
    CHECK_DONE = True

    exit_code, exit_message = 1, "An unknown error occurred"

    #! Add .machine()
    system, release, version, machine = platform.system(), platform.release(), platform.version(), platform.machine()
    full_config = f"{system} {release} {version} ({machine})"

    # Check for incompatible configuration first
    incompatible_state, _, _, _ = _check_os_list(system, release, version, machine, INCOMPATIBLE_OS_LIST)
    # Check for untested but supported configuration
    untested_state, _, _, _ = _check_os_list(system, release, version, machine, UNTESTED_OS_LIST)
    working_state, _, _, _ = _check_os_list(system, release, version, machine, WORKING_OS_LIST)
    if incompatible_state == OSListExitState.Supported:
        exit_code = 1  # Exit code 1 means an error
        exit_message = (
            f"Your current configuration ({full_config}) is incompatible with this program, please try with a "
            f"configuration that is supported ({_format_os_list(UNTESTED_OS_LIST)} / "
            f"{_format_os_list(WORKING_OS_LIST)})"
        )
    elif working_state == OSListExitState.SystemNotSupported:
        exit_code = 1
        exit_message = (
            f"You are currently on {platform.system()}. Please run this on a supported OS "
            f"({', '.join(WORKING_OS_LIST.keys())})."
        )
    elif working_state == OSListExitState.ReleaseOrVersionNotSupported:
        exit_code = 1
        exit_message = (
            f"You are currently on {full_config}. Use a supported release/version for {system} "
            f"({WORKING_OS_LIST[system]})."
        )
    elif working_state == OSListExitState.Supported:
        exit_code = 0
        exit_message = f"{full_config} with Python {sys.version_info[0]}.{sys.version_info[1]} is fully supported."
    elif untested_state == OSListExitState.Supported:
        exit_code = 2  # Exit Code 2 means a warning
        exit_message = (
            f"Your current configuration ({full_config}) is supported by this program, but untested. Consider using a "
            f"tested configuration ({_format_os_list(WORKING_OS_LIST)})"
        )

    if sys.version_info[:2] not in PY_LIST:
        py_versions_strs = [f"{major}.{minor}" for (major, minor) in PY_LIST]
        exit_code, exit_message = 1, (f"You are currently on {'.'.join([str(x) for x in sys.version_info])}. "
                                      f"Please run this using a supported python version ({', '.join(py_versions_strs)}).")

    if exit_code == 1:
        return RuntimeError(exit_message)
    elif exit_code == 2:
        return UserWarning(exit_message)
    return exit_message

def setup() -> None:
    """Setup the app, this does not include checking for compatibility"""
    # Feed information into globals
    global CONFIG_DONE, exported_logs, base_app_dir, old_cwd
    if CONFIG_DONE or not CHECK_DONE:
        return None
    CONFIG_DONE = True
    exported_vars = _configure()
    exported_logs, base_app_dir, old_cwd = (exported_vars["accumulated_logs"], exported_vars["base_app_dir"],
                                            exported_vars["old_cwd"])
    return None

def do(app_info: AppConfig) -> None:
    """Does the three steps configure, check setup at once."""
    configure(app_info)
    err = check()
    if isinstance(err, RuntimeError):
        raise err
    elif isinstance(err, UserWarning):
        warnings.warn(
            str(err),
            type(err),
            stacklevel=2
        )
    else:
        print(err)
    setup()



"""Security links to aplustools"""
from aplustools.security.passwords import (PasswordFilter, PasswordGenerator, SecurePasswordGenerator,
                                           SimplePasswordGenerator)
from aplustools.security.rand import NumPyRandom, SecretsRandom
from aplustools.security.crypto import (Backend as CryptBackend, set_backend as set_crypt_backend,
                                        exceptions as crypt_exceptions)

__all__ = ["PasswordFilter", "PasswordGenerator", "SimplePasswordGenerator", "SecurePasswordGenerator", "NumPyRandom",
           "SecretsRandom", "CryptBackend", "set_crypt_backend", "crypt_exceptions"]



"""Timing links to aplustools"""
from aplustools.package.chronokit import BasicTimer, PreciseTimeDelta, FlexTimer, PreciseTimeFormat

__all__ = ["BasicTimer", "PreciseTimeFormat", "PreciseTimeDelta", "FlexTimer"]



"""Default app configurations"""
from argparse import Namespace as _Ns
import shutil
import os

from ._app import MainClass
from ._default_modules import DefaultLogger, AdvancedLogger, Offloader
from . import io

from collections import abc as _a
import typing as _ty

__all__ = ["DefaultApp", "DefaultAppTUI", "DefaultServerTUI", "DefaultThemedApp"]


class DefaultApp(MainClass):
    def __init__(self, parsed_args: _Ns, logging_level: int, /, setup_thread_pool: bool = False):
        try:
            self.offloader: Offloader | None = None
            if setup_thread_pool:
                self.offloader = Offloader()
        except Exception as e:
            raise Exception("Exception occurred during initialization of the Main class") from e

    def _ensure_offloader(self) -> None:
        if self.offloader is None:
            raise RuntimeError("Pool or/and for loop list is/are not initialized")

    def offload_work(self, task_name: str, task_collection_func: _a.Callable, task: _a.Callable[[], tuple]) -> None:
        """Offloads work to self.offloader"""
        self._ensure_offloader()
        self.offloader.offload_work(task_name, task_collection_func, task)  # type: ignore

    def wait_for_completion(self, task_name: str, /, check_interval: float = 1.0) -> None:
        """Waits for a task with a certain name to complete"""
        self._ensure_offloader()
        self.offloader.wait_for_completion(task_name, check_interval=check_interval)  # type: ignore

    def wait_for_manual_completion(self, task_name: str, /, check_interval: float = 1.0) -> None:
        """Waits manually for a certain task to complete by working down the task results itself"""
        self._ensure_offloader()
        self.offloader.wait_for_manual_completion(task_name, check_interval=check_interval)  # type: ignore

    def timer_tick(self) -> None:
        """Does stuff that needs to be done periodically"""
        if self.offloader is not None:
            self.offloader.tick()

    def close(self) -> None:
        if hasattr(self, "offloader") and self.offloader is not None:
            self.offloader.shutdown()

class DefaultAppTUI(DefaultApp):
    def __init__(self, log_filepath: str | None, parsed_args: _Ns, logging_level: int, /, setup_thread_pool: bool = False) -> None:
        super().__init__(parsed_args, logging_level, setup_thread_pool=setup_thread_pool)
        try:
            self.logger: io.ActLogger = DefaultLogger(log_filepath, logging_level)
            self.system: io.BaseSystemType = io.get_system()
        except Exception as e:
            raise Exception("Exception occurred during initialization of the Main class") from e

    def close(self) -> None:
        if hasattr(self, "logger"):
            sys.stdout = self.logger.restore_pipe(sys.stdout)  # type: ignore
            sys.stderr = self.logger.restore_pipe(sys.stderr)  # type: ignore

    def crash(self, error_title: str, error_text: str, error_description: str) -> bool:
        return self.prompt_user(f"--- {error_title} ---", error_text + "Do you want to restart?", error_description, "error", options=["Y", "N"], default_option="Y") == "Y"

    def prompt_user(self, title: str, message: str, details: str,
                    level: _ty.Literal["debug", "information", "question", "warning", "error"],
                    options: list[str], default_option: str, checkbox_label: str | None = None) -> tuple[str | None, bool]:
        # Log the message at the appropriate level
        log_func = getattr(self.logger, level if level != "information" else "info", self.logger.info)
        log_func(f"{title} | {message}\n{details}")

        terminal_width = shutil.get_terminal_size((80, 20)).columns

        def print_separator():
            print("-" * terminal_width)

        print_separator()
        print(f"\033[1m{title}\033[0m")  # Bold title
        print_separator()
        print(message)
        print()
        if details:
            print(f"\033[2m{details}\033[0m")  # Dimmed details
        print()

        # Display options
        print("Options:")
        for option in options:
            prefix = "-> " if option == default_option else "   "
            print(f"{prefix}{option}")

        selected_option: str | None = None
        checkbox_checked: bool = False

        while selected_option not in options:
            try:
                user_input = input(f"\nSelect option [{default_option}]: ").strip()
                if user_input == "":
                    selected_option = default_option
                elif user_input in options:
                    selected_option = user_input
                else:
                    print(f"Invalid option. Choose one of: {', '.join(options)}")
            except KeyboardInterrupt:
                print("\nCancelled.")
                return None, False

        if checkbox_label:
            while True:
                checkbox_input = input(f"{checkbox_label} [y/N]: ").strip().lower()
                if checkbox_input in ("y", "yes"):
                    checkbox_checked = True
                    break
                elif checkbox_input in ("n", "no", ""):
                    checkbox_checked = False
                    break
                else:
                    print("Please enter y or n.")

        print_separator()
        print(f"Selected: {selected_option} | {'☑' if checkbox_checked else '☐'} {checkbox_label or ''}")
        print_separator()
        return selected_option, checkbox_checked

class DefaultServerTUI(DefaultAppTUI):
    def __init__(self, log_filepath: str, parsed_args: _Ns, logging_level: int, /, always_restart: bool = False,
                 setup_thread_pool: bool = False) -> None:
        super().__init__(log_filepath, parsed_args, logging_level, setup_thread_pool=setup_thread_pool)
        self.always_restart: bool = always_restart

    def prompt_user(self, title: str, message: str, details: str,
                    level: _ty.Literal["debug", "information", "question", "warning", "error"],
                    options: list[str], default_option: str, checkbox_label: str | None = None) -> tuple[str | None, bool]:
        return None, False

    def crash(self, error_title: str, error_text: str, error_description: str) -> bool:
        print(f"--- {error_title} ---\n{error_text}\n{error_description}")
        # for line in (error_text + error_description).split("\n"):
        #     print(line)
        return self.always_restart

class DefaultThemedApp(DefaultApp):
    def __init__(self, logs_directory: str, parsed_args: _Ns, logging_level: int, /, setup_thread_pool: bool = False) -> None:
        super().__init__(parsed_args, logging_level, setup_thread_pool=setup_thread_pool)
        try:
            self.io_manager: io.IOManager = AdvancedLogger(logs_directory, self.prompt_user, logging_level)

            self.system: io.BaseSystemType = io.get_system()
            self.os_theme: io.SystemTheme = self.get_os_theme()
            self.update_theme(self.os_theme)
        except Exception as e:
            raise Exception("Exception occurred during initialization of the Main class") from e

    def open_url(self, url: str) -> None:
        """Opens the given url in a browser"""
        raise NotImplementedError()

    def get_os_theme(self) -> io.SystemTheme:
        """Gets the os theme based on a number of parameters, like environment variables."""
        base = self.system.get_system_theme()
        if not base:
            raw_fallback = str(os.environ.get("DANCER_BACKUP_THEME")).lower()  # Can return None
            fallback = {"light": io.SystemTheme.LIGHT, "dark":io. SystemTheme.DARK}.get(raw_fallback)
            if fallback is None:
                return io.SystemTheme.LIGHT
            return fallback
        return base

    def update_theme(self, new_theme: io.SystemTheme) -> None:
        """Does tasks that need to be done if the theme changes"""
        self.os_theme = new_theme

    def timer_tick(self) -> None:
        super().timer_tick()
        new_theme = self.get_os_theme()
        if new_theme != self.os_theme:
            self.update_theme(new_theme)
        self.io_manager.invoke_prompts()



"""Dancer"""
from . import io, cli, concurrency, data, security, system, timing, web
from ._app import *
from ._default_apps import *
from ._default_modules import *
from .config import do as _do
from .system import diagnose_shutdown_blockers

import typing as _ty
import types as _ts

__version__ = "0.0.1.0b0"

from dataclasses import dataclass as _dc
_DirectoryTree = dict[str, _ty.Union["_DirectoryTree", None]]
@_dc(frozen=True)
class AppConfig:
    """Contains all info on the app"""
    INDEV: bool
    INDEV_KEEP_RUNTIME_FILES: bool
    COMPILE_TIME_FLAGS: bool
    PROGRAM_NAME: str
    PROGRAM_NAME_NORMALIZED: str
    VERSION: int
    VERSION_ADD: str
    WORKING_OS_LIST: dict[str, dict[str, tuple[str, ...]]]
    UNTESTED_OS_LIST: dict[str, dict[str, tuple[str, ...]]]
    INCOMPATIBLE_OS_LIST: dict[str, dict[str, tuple[str, ...]]]
    PY_LIST: list[tuple[int, int]]
    DIR_STRUCTURE: _DirectoryTree
    LOCAL_MODULE_LOCATIONS: list[str]


app_config: AppConfig | None = None
_globs: dict[str, _ty.Any] = {}


def make_global(key: str, value: _ty.Any) -> None:
    global _globs
    _globs[key] = value

def get_global(key: str) -> _ty.Any | None:
    return _globs.get(key)

def configure(configuration: AppConfig):
    ...

def late_init_package(pkg: str, app_info: AppConfig) -> _ts.ModuleType:
    _do(app_info)
    pkg = __import__(pkg, globals(), locals(), [pkg])
    start(pkg.App, pkg.parser, pkg.exit_codes)
    return pkg

def package(pkg_name: str) -> str:
    pkg: _ts.ModuleType = __import__(pkg_name, globals(), locals(), [pkg_name])
    _do(app_info=pkg.app_info)
    pkg_app: _ts.ModuleType = __import__(f"{pkg_name}.app", globals(), locals(), [pkg_name])
    start(pkg_app.App, pkg.parser, pkg_app.exit_codes)
    results: str = diagnose_shutdown_blockers(return_result=True)
    return results



"""Default modules used in apps"""
from dataclasses import dataclass
import logging
import sys
import time
import json
from traceback import format_exc as _format_exc
import os

from packaging.version import Version as _Version, InvalidVersion as _InvalidVersion
import requests
from aplustools.io import SingletonMeta
from aplustools.io.fileio import os_open

from . import config, io

from collections import abc as _a
import typing as _ty

__all__ = ["DefaultLogger", "AdvancedLogger", "Offloader", "UpdateResult", "UpdateChecker", "Translator", "Translation"]


def DefaultLogger(log_filepath: str | None, logging_level: int) -> io.ActLogger:
    """Returns a configured ActLogger instance"""
    # Setup ActLogger
    logger: io.ActLogger
    if log_filepath is not None:
        logger = io.ActLogger(log_to_file=True, filepath=log_filepath)
    else:
        logger = io.ActLogger()
    sys.stdout = logger.create_pipe_redirect(sys.stdout, level=logging.DEBUG)
    sys.stderr = logger.create_pipe_redirect(sys.stderr, level=logging.ERROR)
    mode = getattr(logging, logging.getLevelName(logging_level).upper())
    if mode is not None:
        logger.setLevel(mode)
    for exported_line in config.exported_logs.split("\n"):
        logger.debug(exported_line)  # Flush config prints
    return logger

def AdvancedLogger(logs_directory: str | None, prompt_user: _a.Callable, logging_level: int) -> io.IOManager:
    """Returns a configured IOManager instance, this internally uses ActLogger and ActLogger is Singleton"""
    # Setup IOManager
    io_manager: io.IOManager = io.IOManager()
    if logs_directory is not None:
        io_manager.init(prompt_user, logs_directory, config.INDEV)
    else:
        raise NotImplementedError("Not logging to a directory is not yet supported for IOManager")
    mode = getattr(logging, logging.getLevelName(logging_level).upper())
    if mode is not None:
        io_manager.set_logging_level(mode)
    for exported_line in config.exported_logs.split("\n"):
        io_manager.debug(exported_line)  # Flush config prints
    return io_manager

class Offloader:
    """TBA"""
    def __init__(self) -> None:
        from aplustools.io.concurrency import LazyDynamicThreadPoolExecutor, ThreadSafeList
        self.pool: LazyDynamicThreadPoolExecutor = LazyDynamicThreadPoolExecutor(0, 2, 1.0, 1)
        self._for_loop_list: list[tuple[_ty.Callable[[_ty.Any], _ty.Any], tuple[_ty.Any]]] = ThreadSafeList()
        self._running_tasks: set[str] = set()
        self.max_collections_per_timer_tick: int = 5

    def _check_pool(self) -> bool:
        return not (self.pool is None or self._for_loop_list is None)

    def _ensure_pool(self) -> None:
        if not self._check_pool():
            raise RuntimeError("Pool or/and for loop list is/are not initialized")

    def offload_work(self, task_name: str, task_collection_func: _a.Callable, task: _a.Callable[[], tuple[...]]) -> None:
        self._ensure_pool()
        if task_name in self._running_tasks:
            raise RuntimeError(f"Cannot have two tasks with the name '{task_name}' running at the same time.")
        self._running_tasks.add(task_name)
        self.pool.submit(lambda:
                             self._for_loop_list.append(
                                 (task_name, task_collection_func, task())
                             )
                         )

    def wait_for_completion(self, task_name: str, /, check_interval: float = 1.0) -> None:
        self._ensure_pool()
        while task_name in self._running_tasks:
            time.sleep(check_interval)

    def wait_for_manual_completion(self, task_name: str, /, check_interval: float = 1.0) -> None:
        self._ensure_pool()
        while task_name in self._running_tasks:
            time.sleep(check_interval)
            if self._for_loop_list:
                entry = self._for_loop_list.pop()
                name, func, args = entry
                func(*args)
                self._running_tasks.remove(name)

    def tick(self) -> None:
        ...

    def shutdown(self, wait: bool = True, *, cancel_futures: bool = False) -> None:
        self.pool.shutdown(wait, cancel_futures=cancel_futures)

@dataclass  # TODO:
class UpdateResult:
    ...

class UpdateChecker:
    # Flags
    INFORM_ABOUT_UPDATE_INFO_FORMAT: bool = True
    CHECK_FOR_UPDATE: bool = True
    UPDATE_CHECK_REQUEST_TIMEOUT: float = 4.0
    SHOW_UPDATE_TIMEOUT: bool = False
    SHOW_UPDATE_ERROR: bool = False
    SHOW_UPDATE_INFO: bool = True
    SHOW_NO_UPDATE_INFO: bool = False
    def __init__(self, update_check_url: str):
        raise NotImplementedError()

    def get_update_result(self) -> tuple[bool, tuple[str, str, str, str], tuple[str | None, tuple[str, str]], tuple[list[str], str], _a.Callable[[str], _ty.Any]]:
        """
        Checks for an update and returns the result.
        """
        icon: str = "Information"
        title: str = "Title"
        text: str = "Text"
        description: str = "Description"
        checkbox: str | None = None
        checkbox_setting: tuple[str, str] = ("", "")
        standard_buttons: list[str] = ["Ok"]
        default_button: str = "Ok"
        retval_func: _a.Callable[[str], _ty.Any] = lambda button: None
        do_popup: bool = True

        try:  # Get update content
            response: requests.Response = requests.get(
                self.update_check_url,
                timeout=float(self.UPDATE_CHECK_REQUEST_TIMEOUT))
        except requests.exceptions.Timeout:
            title, text, description = "Update Info", ("The request timed out.\n"
                                                       "Please check your internet connection, "
                                                       "and try again."), _format_exc()
            standard_buttons, default_button = ["Ok"], "Ok"
            checkbox, checkbox_setting = "Do not show again", ("auto", "show_update_timeout")
            show_update_timeout: bool = self.SHOW_UPDATE_TIMEOUT
            if not show_update_timeout:
                do_popup = False
            return (do_popup,
                    (icon, title, text, description),
                    (checkbox, checkbox_setting),
                    (standard_buttons, default_button), retval_func)
        except requests.exceptions.RequestException:
            title, text, description = "Update Info", ("There was an error with the request.\n"
                                                       "Please check your internet connection and antivirus, "
                                                       "and try again."), _format_exc()
            standard_buttons, default_button = ["Ok"], "Ok"
            return (do_popup,
                    (icon, title, text, description),
                    (checkbox, checkbox_setting),
                    (standard_buttons, default_button), retval_func)
        except Exception as e:
            return (self.SHOW_UPDATE_ERROR,
                    ("Warning", "Update check failed", "Due to an internal error,\nthe operation could not be completed.", _format_exc()),
                    ("Do not show again", ("auto", "show_update_error")),
                    (["Ok"], "Ok"), lambda button: None)

        try:  # Parse update content
            update_json: dict = response.json()
            current_version = _Version(f"{config.VERSION}{config.VERSION_ADD}")
            found_version: _Version | None = None
            found_release: dict | None = None
            found_push: bool = False

            for release in update_json["versions"]:
                release_version = _Version(release["versionNumber"])
                if release_version == current_version:
                    found_version = release_version
                    found_release = release
                    found_push = False  # Doesn't need to be set again
                if release_version > current_version:
                    push = release["push"].title() == "True"
                    if found_version is None or (release_version > found_version and push):
                        found_version = release_version
                        found_release = release
                        found_push = push
                # if found_release and found_version != current_version:
                #     raise NotImplementedError
        except (requests.exceptions.JSONDecodeError, _InvalidVersion, NotImplementedError):
            icon = "Information"  # Reset everything to default, we don't know when the error happened
            title, text, description = "Update Info", "There was an error when decoding the update info.", _format_exc()
            checkbox, checkbox_setting = None, ("", "")
            standard_buttons, default_button = ["Ok"], "Ok"
            retval_func = lambda button: None
            return (do_popup,
                    (icon, title, text, description),
                    (checkbox, checkbox_setting),
                    (standard_buttons, default_button), retval_func)

        show_update_info: bool = self.SHOW_UPDATE_INFO
        show_no_update_info: bool = self.SHOW_NO_UPDATE_INFO

        if found_version != current_version and show_update_info and found_push:
            title = "There is an update available"
            text = (f"There is a newer version ({found_version}) "
                    f"available.\nDo you want to open the link to the update?")
            description = str(found_release.get("description"))  # type: ignore
            checkbox, checkbox_setting = "Do not show again", ("auto", "show_update_info")
            standard_buttons, default_button = ["Yes", "No"], "Yes"

            def retval_func(button: str) -> None:
                """TBA"""
                if button == "Yes":
                    url = str(found_release.get("updateUrl", "None"))  # type: ignore
                    if url.title() == "None":
                        link = update_json["metadata"].get("sorryUrl", "https://example.com")
                    else:
                        link = url
                    self.open_url(link)
        elif show_no_update_info and found_version <= current_version:
            title = "Update Info"
            text = (f"No new updates available.\nChecklist last updated "
                    f"{update_json['metadata']['lastUpdated'].replace('-', '.')}.")
            description = f" --- v{found_version} --- \n{found_release.get('description')}"  # type: ignore
            checkbox, checkbox_setting = "Do not show again", ("auto", "show_no_update_info")
        elif show_no_update_info and not found_push:
            title = "Info"
            text = (f"New version available, but not recommended {found_version}.\n"
                    f"Checklist last updated {update_json['metadata']['lastUpdated'].replace('-', '.')}.")
            description = str(found_release.get("description"))  # type: ignore
            checkbox, checkbox_setting = "Do not show again", ("auto", "show_no_update_info")
        else:
            title, text, description = "Update Info", "There was a logic-error when checking for updates.", ""
            do_popup = False
        return (do_popup,
                (icon, title, text, description),
                (checkbox, checkbox_setting),
                (standard_buttons, default_button), retval_func)

    def show_update_result(self, update_result: tuple[bool, tuple[str, str, str, str], tuple[str | None, tuple[str, str]], tuple[list[str], str], _a.Callable[[str], _ty.Any]]) -> None:
        """
        Shows update result using a message box
        """
        (do_popup,
         (icon, title, text, description),
         (checkbox, checkbox_setting),
         (standard_buttons, default_button), retval_func) = update_result
        if do_popup:
            retval, checkbox_checked = self.prompt_user(title, text, description, icon, standard_buttons,
                                                        default_button, checkbox)
            retval_func(retval)
            if checkbox is not None and checkbox_checked:
                setattr(self, checkbox_setting[1].upper(), False)


            # self.update_check_url: str = update_check_url  # TODO: Create class UpdateChecker
            # if self.INFORM_ABOUT_UPDATE_INFO_FORMAT:
            #     print("INFORMATION ABOUT UPDATE INFO FORMAT:: https://raw.githubusercontent.com/Giesbrt/Automaten/main/meta/update_check.json")
            # if self.CHECK_FOR_UPDATE:
            # result = self.get_update_result()
            # self.show_update_result(result)

class Translation(metaclass=SingletonMeta): ...

class Translator(metaclass=SingletonMeta):
    def __init__(self, localisation_dir: str) -> None:
        self.localisation_dir: str = localisation_dir
        self.localisations: dict[str, str] = {}
        self._tr: _ty.Type[Translation] | None = None
        self.loaded_localisation: str = ""
        self.reload_localisations()

    def reload_localisations(self) -> None:
        self.localisations.clear()
        for file in os.listdir(self.localisation_dir):
            self.localisations[file.removesuffix(".json")] = file
        return None

    def set_translation_cls(self, translation_cls: _ty.Type[Translation]) -> None:
        if not issubclass(translation_cls, Translation):
            raise ValueError(f"The parameter translation_cls needs to be of type Translation")
        self._tr = translation_cls

    def check_localisations(self) -> None:
        current_localisation: str = self.loaded_localisation
        for localisation in self.localisations.keys():
            self.load(localisation)
        self.load(current_localisation)
        return None

    def load(self, localisation: str) -> None:
        file: str | None = self.localisations.get(localisation)
        if file is None:
            raise ValueError(f"The localisation for '{localisation}' is not currently loaded")
        filepath: str = os.path.join(self.localisation_dir, file)

        translation: dict[str, str]
        with os_open(filepath, "r") as f:
            try:
                translation = json.loads(f.read())
            except json.JSONDecodeError as e:
                raise RuntimeError(f"Failed to parse localisation file '{filepath}'") from e

        if self._tr is None:
            raise RuntimeError("You need to set self.tr before you can load a translation. "
                               "Tr exists to enable type hinting.")

        keys: set[str] = set()
        keys.update(dir(self._tr))
        keys.update(self._tr.__annotations__.keys())

        for key in keys:
            if not key.startswith("_"):
                value: str | None = translation.get(key)
                if value is None:
                    raise RuntimeError(f"The key '{key}' was not provided by the translation for '{localisation}'")
                setattr(self._tr, key, value)
        self.loaded_localisation = localisation
        return None



"""Storage links to aplustools"""
from aplustools.data import (SingletonMeta, Sorters, what_module, what_class, what_func, unnest_iterable,
                             cutoff_iterable, cutoff_string, minmax, beautify_json, format_type, align_to_next,
                             align_to_previous)
from aplustools.data.bintools import (encode_float, decode_float, encode_integer, decode_integer, bit_length,
                                      bytes_length, get_variable_bytes_like, read_variable_bytes_like,
                                      expected_varint_length_bytecount, BitBuffer, set_bits, bits,
                                      bytes_to_human_readable_binary_iec, bytes_to_human_readable_decimal_si,
                                      bits_to_human_readable)
from aplustools.data.dummy import Dummy3 as Dummy
from aplustools.data.storage import JSONStorage, SQLite3Storage, BinaryStorage

__all__ = ["encode_float", "decode_float", "encode_integer", "decode_integer", "bit_length", "bytes_length",
           "get_variable_bytes_like", "read_variable_bytes_like", "expected_varint_length_bytecount", "BitBuffer",
           "set_bits", "bits", "bytes_to_human_readable_binary_iec", "bytes_to_human_readable_decimal_si",
           "bits_to_human_readable", "Dummy", "JSONStorage", "SQLite3Storage", "BinaryStorage", "SingletonMeta",
           "Sorters", "what_module", "what_class", "what_func", "unnest_iterable", "cutoff_iterable", "cutoff_string",
           "minmax", "beautify_json", "format_type", "align_to_next", "align_to_previous", "Singleton"]


class Singleton(metaclass=SingletonMeta): ...



"""Known styles and themes"""
import os.path
from importlib.resources import files as _files

styles: dict[str, str] = {}
base_styles_path = _files("dancer._styling.styles")
for styles_file in base_styles_path.iterdir():
    if styles_file.name.endswith(".qst"):
        style_content: str
        with styles_file.open("r") as f:
            style_content = f.read()
        styles[os.path.basename(str(styles_file))] = style_content

themes: dict[str, str] = {}
base_themes_path = _files("dancer._styling.themes")
for themes_file in base_themes_path.iterdir():
    if themes_file.name.endswith(".qth"):
        theme_content: str
        with themes_file.open("r") as f:
            theme_content = f.read()
        themes[os.path.basename(str(themes_file))] = theme_content



"""Qt-Styling module"""
from string import Template, ascii_letters, digits
import os
import re

from PySide6.QtWidgets import QWidget, QMainWindow
from PySide6.QtGui import QPalette, QColor
from PySide6.QtCore import Qt, QObject
from PySide6 import QtWidgets as _QtWidgets, QtGui as _QtGui, QtCore as _QtCore

from aplustools.io.fileio import os_open

# Standard typing imports for aps
import collections.abc as _a
import typing as _ty
import types as _ts

from ._known_styling import styles as known_styles, themes as known_themes


def assign_object_names_iterative(parent: QObject, prefix: str = "", exclude_primitives: bool = True) -> None:
    """Assign object names iteratively to children using a hierarchy."""
    stack: list[tuple[QObject, str]] = [(parent, prefix)]  # Stack to manage widgets and their prefixes

    # List of primitive classes to exclude if `exclude_primitives` is True
    primitives: tuple[QObject, ...] = ()

    while stack:
        current_parent, current_prefix = stack.pop()

        parent_var_names = {
            obj: name
            for name, obj in vars(current_parent).items()
            if isinstance(obj, current_parent.__class__) or obj in current_parent.children()
        }

        for child in current_parent.children():
            # Construct the object name using colons for hierarchy
            var_name = parent_var_names.get(child, child.__class__.__name__)
            object_name = f"{current_prefix}-{var_name}" if current_prefix else var_name

            # Assign the object name
            if hasattr(child, "setObjectName"):
                child.setObjectName(object_name)
                # print(object_name, child)

            # Check if we should process this child further
            if not exclude_primitives or not isinstance(child, primitives):
                stack.append((child, object_name))

class AbstractMainWindow(_QtWidgets.QMainWindow):
    default_style: str

    def setup_gui(self) -> None:
        """
        Configure the main graphical user interface (GUI) elements of the application.

        This method sets up various widgets, layouts, and configurations required for the
        main window interface. It is called after initialization and prepares the interface
        for user interaction.

        Note:
            This method is intended to be overridden by subclasses with application-specific
            GUI components.
        """
        raise NotImplementedError

    def set_window_icon(self, absolute_path_to_icon: str) -> None:
        self.setWindowIcon(_QtGui.QIcon(absolute_path_to_icon))

    def set_window_title(self, title: str) -> None:
        self.setWindowTitle(title)

    def set_window_geometry(self, x: int, y: int, height: int, width: int) -> None:
        self.setGeometry(_QtCore.QRect(x, y, width, height))

    def set_window_dimensions(self, height: int, width: int) -> None:
        self.resize(_QtCore.QSize(width, height))

    def set_font(self, font_str: str) -> None:
        font = _QtGui.QFont(font_str)
        self.setFont(font)
        for child in self.findChildren(QWidget):
            child.setFont(font)
        self.update()
        self.repaint()

    def set_theme_to_singular(self, theme_str: str, widget_or_window: QWidget) -> None:
        """Applies a theme string to a singular object"""
        widget_or_window.setStyleSheet(theme_str)

    def set_global_theme(self, theme_str: str, base: str | None = None) -> None:
        self.setStyleSheet(theme_str)
        if base is not None:
            if not hasattr(self, "default_style"):
                self.default_style = self.app.style().objectName()
            self.app.setStyle(base)
        else:
            if hasattr(self, "default_style"):
                self.app.setStyle(self.default_style)

    def internal_obj(self) -> QMainWindow:
        return self

    def start(self) -> None:
        self.show()
        self.raise_()

    def close(self) -> None:
        QMainWindow.close(self)

class AppStyle:
    """QApp Styles"""
    Windows11 = "windows11"
    WindowsVista = "windowsvista"
    Windows = "Windows"
    Fusion = "Fusion"
    Default = None

class Style:
    _loaded_styles: dict[str, _ty.Self] = {}

    def __init__(self, style_name: str, for_paths: list[str], parameters: list[str],
                 palette_parameter: list[str]) -> None:
        self._style_name: str = style_name
        self._for_paths: list[str] = for_paths
        self._parameters: list[str] = parameters
        self._palette_parameter: list[str] = palette_parameter
        self._loaded_styles[style_name] = self

    def get_style_name(self) -> str:
        return self._style_name

    def get_parameters(self) -> list[str]:
        return self._parameters

    def get_palette_parameters(self) -> list[str]:
        return self._palette_parameter

    def get_for_paths(self) -> list[str]:
        return self._for_paths.copy()

    @classmethod
    def get_loaded_style(cls, style_name: str, for_theme: "Theme" | _ty.Literal["*"]) -> _ty.Self | None:
        possible_style = cls._loaded_styles.get(style_name)
        if possible_style is None:
            return None
        if for_theme == "*" or for_theme.is_compatible(possible_style):
            return possible_style
        return None

    @classmethod
    def get_loaded_styles(cls, for_theme: "Theme") -> list[_ty.Self]:
        possible_styles = cls._loaded_styles.values()
        found_styles: list[_ty.Self] = []
        for possible_style in possible_styles:
            if for_theme.is_compatible(possible_style):
                found_styles.append(possible_style)
        return found_styles

    @classmethod
    def load_from_file(cls, filepath: str) -> _ty.Self:
        with os_open(filepath, "r") as f:
            content = f.read()
        filename = os.path.basename(filepath)
        return cls.load_from_content(filename, content.decode("utf-8"))

    @staticmethod
    def _parse_paths(input_str: str) -> list[str]:
        # Step 1: Extract the part after "for" and before the ";"
        match = re.match(r'for\s?(.+);', input_str.strip())
        if not match:
            raise ValueError("Input must start with 'for' and end with ';'")

        content = match.group(1).strip()

        # Step 2: Recursive function to expand curly braces
        def expand(segment):
            if "{" not in segment:
                return [segment]

            # Find the first set of braces
            start = segment.index("{")
            end = segment.index("}", start)

            # Before the braces, the braces content, and after the braces
            prefix = segment[:start]
            brace_content = segment[start + 1:end]
            suffix = segment[end + 1:]

            # Expand the content inside the braces
            options = brace_content.split(",")

            # Recursively expand the rest of the string
            expanded_suffixes = expand(suffix)

            # Combine each option with the expanded suffixes
            result = []
            for option in options:
                for expanded_suffix in expanded_suffixes:
                    result.append(f"{prefix}{option.strip()}{expanded_suffix}")
            return result

        # Step 3: Call the recursive function on the cleaned-up string
        return expand(content)

    @classmethod
    def load_from_content(cls, filename: str, content: str) -> _ty.Self:
        """TBA"""
        if filename.endswith(".qst"):
            style_name: str = os.path.splitext(filename)[0].replace("_", " ").title()
        else:
            raise ValueError(f"Invalid .qst file name: {filename}")

        translation_table = str.maketrans("", "", "\n\t ")
        cleaned_content: str = re.sub(r"//.*?$|/\*.*?\*/", "", content, flags=re.DOTALL | re.MULTILINE)

        if not cleaned_content.startswith("for "):
            raise RuntimeError("Style does not include for directive")

        trans_content: str = cleaned_content.translate(translation_table)
        if trans_content == "":
            raise ValueError("The .qst file is empty.")

        for_line: str
        other_content: str
        for_line, other_content = trans_content.split(";", maxsplit=1)
        for_paths = cls._parse_paths(for_line + ";")

        parameters: list[str] = []
        palette_parameter: list[str] = []
        palette_part = False
        for part in other_content.split(";"):
            if not part:
                continue
            if palette_part:
                if part == "]":
                    palette_part = False
                    continue
                palette_parameter.append(part)
            elif part.startswith("QPalette["):
                palette_part = True
                palette_parameter.append(part.removeprefix("QPalette["))
            else:
                parameters.append(part)

        if palette_part:
            raise RuntimeError("Unterminated QPalette declaration")

        return cls(style_name, for_paths, parameters, palette_parameter)

    @classmethod
    def clear_loaded_styles(cls) -> None:
        cls._loaded_styles.clear()

    def __repr__(self) -> str:
        return (f"Style(style_name={self._style_name}, for_paths={self._for_paths}, parameters={self._parameters}, "
                f"palette_parameter={self._palette_parameter})")

class Theme:
    _loaded_themes: dict[str, _ty.Self] = {}

    def __init__(self, author: str, theme_name: str, theme_str: str, base: str | None, placeholders: list[str],
                 compatible_styling: str | None, load_styles_for: str,
                 inherit_extend_from: tuple[str | None, str | None]) -> None:
        self._author: str = author
        self._theme_name: str = theme_name
        self._theme_uid: str = f"{self._author}::{self._theme_name}"
        self._theme_str: str = theme_str
        self._base: str | None = base
        self._placeholders: list[str] = placeholders
        self._compatible_styling: str | None = compatible_styling
        self._load_styles_for: str = load_styles_for
        self._inherit_extend_from: tuple[str, str] = inherit_extend_from
        self._loaded_themes[self._theme_uid] = self

    @staticmethod
    def _find_special_sequence(s: str) -> tuple[str, str, str]:
        """
        Finds the first non-alphanumeric or non-underscore character in a string,
        collects it and all contiguous characters of the same type immediately following it,
        and returns the resulting substring.

        :param s: The input string to search.
        :return: The substring of contiguous special characters or an empty string if none found.
        """
        allowed_chars = ascii_letters + digits + '_'
        front = ""
        back = ""
        for i, char in enumerate(s):
            if char not in allowed_chars:
                special_sequence = char

                for j in range(i + 1, len(s)):
                    if s[j] not in allowed_chars:
                        special_sequence += s[j]
                    else:
                        back = s[j:]
                        break
                return front, special_sequence, back
            else:
                front += char
        return s, "", ""  # Return empty string if no special characters found

    @staticmethod
    def _to_camel_case(s: str) -> str:
        """
        Converts PascalCase or TitleCase to camelCase.

        :param s: The input string to convert.
        :return: The camel case version of the string.
        """
        return s
        if not s:
            return s
        return s[0].lower() + s[1:]

    def get_theme_uid(self) -> str:
        return self._theme_uid

    def is_theme(self, theme_uid: str) -> bool:
        return self._theme_uid == theme_uid

    @classmethod
    def get_loaded_theme(cls, theme_uid: str) -> _ty.Self | None:
        if theme_uid not in cls._loaded_themes:
            return None
        return cls._loaded_themes[theme_uid]

    def get_base_styling(self) -> str:
        return self._base

    def supports_styles(self) -> bool:
        return self._compatible_styling in ("*", "os")

    def is_compatible(self, style: Style) -> bool:
        load_st_author, load_st_theme = self._load_styles_for.split("::")
        for path in style.get_for_paths():
            author, theme_name, styling, maybe_default, *_ = path.split("::", maxsplit=3) + [""]
            if (author == load_st_author or author == "*") and (theme_name == load_st_theme or theme_name == "*"):
                return True
        return False

    def get_compatible_styles(self) -> list[Style]:
        if not self.supports_styles():
            raise RuntimeError(f"The theme '{self._theme_name}' doesn't support styles")
        return Style.get_loaded_styles(for_theme=self)

    def get_compatible_style(self, name: str) -> Style | None:
        if not self.supports_styles():
            raise RuntimeError(f"The theme '{self._theme_name}' doesn't support styles")
        return Style.get_loaded_style(name, for_theme=self)

    def assemble_qss_placeholder_row(self, placeholders: list) -> str:
        mode, from_theme = self._inherit_extend_from

        if mode is None:
            placeholders.extend(self._placeholders)
            return self._theme_str
        elif mode not in ("inheriting", "extending"):
            raise RuntimeError(f"Unsupported mode '{mode}'")

        theme = self._loaded_themes.get(from_theme)
        if theme is None:
            raise RuntimeError(f"Unknown theme '{from_theme}'")
        if mode == "inheriting":
            placeholders.extend(self._placeholders)
        result = theme.assemble_qss_placeholder_row(placeholders) + self._theme_str
        if mode == "extending":
            placeholders.extend(self._placeholders)
        return result

    def apply_style(self, style: Style, palette: QPalette,
                    transparency_mode: _ty.Literal["none", "author", "direct", "indirect"] = "none"
                    ) -> tuple[str, QPalette]:
        # TODO: Make transparency mode, make everything better
        if transparency_mode != "none":
            raise NotImplementedError("Transparency modes are not supported yet")
        if not self.is_compatible(style):  # Remove ?
            raise RuntimeError()
        placeholders: list[str] = []
        raw_qss = Template(self.assemble_qss_placeholder_row(placeholders))

        formatted_placeholder: dict[str, str] = {}
        if style is not None:
            for style_placeholder in style.get_parameters():  # Move style placeholders and QPalette conversion up
                front, back = [c.strip() for c in style_placeholder.split(":")]
                formatted_placeholder[front] = back
            for qpalette_placeholder in style.get_palette_parameters():
                key, val = qpalette_placeholder.split(":")
                color: QColor
                if val.startswith("#") and val[1:].isalnum():
                    color = QColor.fromString(val)
                elif (val.startswith("rba(") or val.startswith("rgba(")) and val.endswith(")"):
                    pattern = r"rgba?\((\d+),\s*(\d+),\s*(\d+)(?:,\s*(\d+))?\)"
                    match = re.match(pattern, val)
                    if match:
                        r, g, b = map(int, match.groups()[:3])  # Extract R, G, B
                        a = int(match.group(4)) if match.group(4) else 255  # Extract A or default to 255 (fully opaque)
                        color = QColor(r, g, b, a)  # Create QColor object
                    else:
                        raise ValueError(f"Invalid color format in style {style.get_style_name()}'s QPalette: {val}")
                else:
                    raise ValueError(f"Invalid color format in style {style.get_style_name()}'s QPalette: {val}")
                palette.setColor(getattr(QPalette.ColorRole, key), color)

        for placeholder in placeholders:
            front, assignment_type, end = self._find_special_sequence(placeholder)
            if assignment_type == "~=":
                if end.startswith("QPalette."):
                    placeholder = palette.color(getattr(QPalette.ColorRole, end.removeprefix("QPalette."))).name()
                elif end.startswith("#") and end[1:].isalnum():
                    placeholder = end
                elif (end.startswith("rba(") or end.startswith("rgba(")) and end.endswith(")"):
                    placeholder = end
                elif end.startswith("url(") and end.endswith(")"):
                    placeholder = end
                    if front not in formatted_placeholder:
                        formatted_placeholder[front] = placeholder
                    continue
                else:
                    placeholder = QColor(getattr(Qt.GlobalColor, self._to_camel_case(end)))
                if not isinstance(placeholder, str):
                    placeholder = f"rgba({placeholder.red()}, {placeholder.green()}, {placeholder.blue()}, {placeholder.alpha()})"
                if front not in formatted_placeholder:
                    formatted_placeholder[front] = placeholder
            elif assignment_type == "==":
                if end.startswith("QPalette."):
                    placeholder = palette.color(getattr(QPalette.ColorRole, end.removeprefix("QPalette."))).name()
                elif end.startswith("#") and end[1:].isalnum():
                    placeholder = end
                elif (end.startswith("rba(") or end.startswith("rgba(")) and end.endswith(")"):
                    placeholder = end
                elif end.startswith("url(") and end.endswith(")"):
                    placeholder = end
                    formatted_placeholder[front] = placeholder
                    continue
                else:
                    placeholder = QColor(getattr(Qt.GlobalColor, self._to_camel_case(end)))
                if not isinstance(placeholder, str):
                    placeholder = f"rgba({placeholder.red()}, {placeholder.green()}, {placeholder.blue()}, {placeholder.alpha()})"
                formatted_placeholder[front] = placeholder
            else:
                raise RuntimeError("Malformed placeholder")

        formatted_qss = raw_qss.safe_substitute(**formatted_placeholder)
        return formatted_qss, palette

    @classmethod
    def load_from_file(cls, filepath: str) -> _ty.Self:
        with os_open(filepath, "r") as f:
            content = f.read()
        filename = os.path.basename(filepath)
        return cls.load_from_content(filename, content.decode("utf-8"))

    @classmethod
    def load_from_content(cls, filename: str, content: str) -> _ty.Self:
        if "_" in filename and filename.endswith(".qth"):
            author, theme_name_ext = filename.split("_", 1)
            theme_name = os.path.splitext(theme_name_ext)[0]  # Remove ".qth"
        else:
            raise ValueError(f"Invalid .qth file name: {filename}")
        # TODO: Use string translation here
        cleaned_content = re.sub(r"//.*?$|/\*.*?\*/", "", content, flags=re.DOTALL | re.MULTILINE).strip() + "\n"
        if cleaned_content == "":
            raise ValueError("The .qth file is empty.")

        mode: _ty.Literal["extending", "inheriting"] | None = None
        from_theme: str | None = None
        if cleaned_content.startswith("inheriting") or cleaned_content.startswith("extending"):
            mode_line_s, cleaned_content = cleaned_content.split(";", maxsplit=1)
            mode, from_theme = mode_line_s.split(" ", 1)

        config_line, other_content = cleaned_content.lstrip().split("\n", maxsplit=1)
        style_metadata = config_line.split("/")
        # print("Config Line:  ", repr(config_line), style_metadata)
        if len(style_metadata) < 3:
            raise ValueError(f"The config line of the .qth file is invalid: '{config_line}'")
        base_app_style = style_metadata[0] if len(style_metadata[0].strip()) > 0 else None
        compatible_styling = style_metadata[1] if len(style_metadata[1].strip()) > 0 else None
        style_precautions = style_metadata[2] if len(style_metadata[2].strip()) > 0 else None
        print(f"Discovered Mode+Style ({author}::{theme_name}): '{mode} {from_theme}'; '{base_app_style, compatible_styling, style_precautions}'")

        lines = other_content.split("\n")

        qss: str = ""
        raw_placeholders: list[str] = []
        for i, line in enumerate(lines):
            if line.startswith("ph:"):
                raw_placeholders.extend(line.removeprefix("ph:").split(";"))
                raw_placeholders.extend(''.join(lines[i+1:]).split(";"))
                break
            else:
                qss += line

        placeholders: list[str] = []
        for raw_placeholder in raw_placeholders:
            placeholder = raw_placeholder.strip()
            if placeholder != "":
                placeholders.append(placeholder)
        # TODO: add other attributes more clearly
        load_styles_for = from_theme if style_precautions == "reuse_st" and from_theme is not None else f"{author}::{theme_name}"
        inherit_extend = (mode, from_theme)
        return cls(author, theme_name, qss.strip(), base_app_style, placeholders, compatible_styling, load_styles_for,
                   inherit_extend)

    @classmethod
    def clear_loaded_themes(cls) -> None:
        cls._loaded_themes.clear()

    def __repr__(self) -> str:
        return (f"Theme(theme_uid={self._theme_uid}, theme_str={self._theme_str[:10]}, base={self._base}, "
                f"placeholder={self._placeholders[:3]}, compatible_styling={self._compatible_styling}, "
                f"load_styles_for={self._load_styles_for}, inherit_extend_from={self._inherit_extend_from})")



"""Everything around the base app"""
from argparse import ArgumentParser as _Ag, Namespace as _Ns
from traceback import format_exc as _format_exc
import logging
import sys
import os

from . import config

from collections import abc as _a
import typing as _ty


class MainClass(_ty.Protocol):
    """
    The main class that gets executed in the application lifecycle.

    This protocol defines the core structure for running an application, handling execution,
    cleanup, and crash-related behavior.
    """
    def __init__(self, parsed_args: _Ns, logging_level: int) -> None: ...
    def exec(self) -> int:
        """
        Execute the application logic.

        This method blocks until the application is complete.

        Returns:
            int: An integer error code indicating the exit status of the application.
        """
        raise NotImplementedError()
    def close(self) -> None:
        """
        Clean up resources after application execution or a crash.

        This will be called after `exec()` completes or a crash occurs to ensure
        all resources are properly released.
        """
        raise NotImplementedError()
    def crash(self, error_title: str, error_text: str, error_description: str) -> bool:
        """
        Handle application crash behavior.

        This method is called after a crash occurs to display or log crash-related
        information. It typically prompts the user whether to restart the application.

        Args:
            error_title (str): A short title summarizing the error.
            error_text (str): A brief message describing the error.
            error_description (str): A detailed explanation of the error.

        Returns:
            bool: True if the application should restart; False otherwise.
        """
        raise NotImplementedError()
    def prompt_user(self, title: str, message: str, details: str,
                    level: _ty.Literal["debug", "information", "question", "warning", "error"],
                    options: list[str], default_option: str, checkbox_label: str | None = None) -> tuple[str | None, bool]:
        """
        Display a prompt to the user with a message, optional checkbox, and buttons to choose from.

        This method can be used to display critical information, questions, or warnings, and
        allows the user to choose a predefined action. An optional checkbox can also be displayed
        (e.g., "Don't show this again").

        Args:
            title (str): The window or dialog title.
            message (str): A short message to show to the user.
            details (str): A more detailed description to accompany the message.
            level (Literal): The level of the user action (e.g., "information", "warning", "question", etc.).
            options (list[str]): A list of button labels representing user actions.
            default_option (str): The default button/action that will be preselected.
            checkbox_label (str | None): The label for an optional checkbox; if None, no checkbox is shown.

        Returns:
            tuple[str | None, bool]: A tuple containing the selected action (or None if no selection was made),
            and a boolean indicating whether the checkbox was activated (True) or not (False).
        """
        raise NotImplementedError()

def start(main_class: _ty.Type[MainClass], arg_parser: _Ag | None = None, exit_codes: dict[int, _a.Callable[[], None]] | None = None) -> None:
    """Starts the app and handles error catching"""
    if exit_codes is None:
        sys.argv[0] = os.path.join(config.old_cwd, sys.argv[0])
        if config.is_compiled():
            exit_codes = {
                1000: lambda: os.execv(sys.executable, [sys.executable] + sys.argv[1:])  # RESTART_CODE (only works compiled)
            }
        else:
            exit_codes = {
                1000: lambda: os.execv(sys.executable, [sys.executable] + sys.argv)  # RESTART_CODE (only works uncompiled)
            }
    if arg_parser is None:
        arg_parser = _Ag(description=f"{config.PROGRAM_NAME}")
    dp_app: MainClass | None = None
    current_exit_code: int = -1

    arg_parser.add_argument("--logging-level", choices=["DEBUG", "INFO", "WARN", "WARNING", "ERROR"], default="INFO",
                        help="Logging level (default: INFO)")
    arg_parser.add_argument("--version", action="store_true", help="Shows the version of the program and dancer")
    args = arg_parser.parse_args()

    if args.version:
        from . import __version__
        print(f"Dancer {__version__} running {config.PROGRAM_NAME} {config.get_version_str()}")
        return

    logging_level_str: str = args.logging_level
    logging_level: int
    input_logging_level = getattr(logging, logging_level_str.upper(), None)
    if input_logging_level is None:
        logging.error(f"Invalid logging mode: {logging_level_str}")
        logging_level = logging.INFO
    else:
        logging_level = input_logging_level

    if config.INDEV:
        print("Setting logging level to debug because INDEV flag is set ...")
        logging_level = logging.DEBUG
    print(f"Starting {config.PROGRAM_NAME} {str(config.VERSION) + config.VERSION_ADD} with py{'.'.join([str(x) for x in sys.version_info])} {'[INDEV]' if config.INDEV else ''} ...")
    try:
        dp_app = main_class(args, logging_level)
        current_exit_code = dp_app.exec()
    except Exception as e:
        perm_error = False
        if isinstance(e.__cause__, PermissionError):
            perm_error = True
        if perm_error:
            error_title = "Warning"
            error_text = (f"{config.PROGRAM_NAME} encountered a permission error. This error is unrecoverable.     \n"
                          "Make sure no other instance is running and that no internal app files are open.     ")
        else:
            error_title = "Fatal Error"
            error_text = (f"There was an error while running the app {config.PROGRAM_NAME}.\n"
                          "This error is unrecoverable.\n"
                          "Please submit the details to our GitHub issues page.")
        error_description = _format_exc()

        if dp_app is not None:
            should_restart: bool = dp_app.crash(error_title, error_text, error_description)
            if should_restart:
                current_exit_code = 1000

        logger: logging.Logger = logging.getLogger("ActLogger")
        if not logger.hasHandlers():
            print(error_description.strip())  # We print, in case the logger is not initialized yet
        else:
            for line in error_description.strip().split("\n"):
                logger.error(line)
    finally:
        if dp_app is not None:
            dp_app.close()
        # results: str = diagnose_shutdown_blockers(return_result=True)
        exit_codes.get(current_exit_code, lambda: sys.exit(current_exit_code))()



"""(auto) CLI links to aplustools"""
from aplustools.package.autocli import ArgStructBuilder, ArgumentParsingError, Argumint, EndPoint, NoDefault

__all__ = ["ArgStructBuilder", "ArgumentParsingError", "Argumint", "EndPoint", "NoDefault"]
